import logging
import os
from unittest import mock
from lxml import etree

from pcs_test.tools.assertions import AssertPcsMixin
from pcs_test.tools.custom_mock import MockLibraryReportProcessor
from pcs_test.tools.misc import (
    get_test_resource,
    get_tmp_file,
    write_file_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunner
from pcs_test.tools.xml import etree_to_str

from pcs import settings
from pcs.lib.external import CommandRunner

# pylint: disable=line-too-long


class CachedCibFixture(AssertPcsMixin):
    def __init__(self, cache_name, empty_cib_path):
        self._empty_cib_path = empty_cib_path
        self._cache_name = cache_name
        self._cache_path = None
        self._pcs_runner = None

    def _setup_cib(self):
        raise NotImplementedError()

    def set_up(self):
        fixture_dir = get_test_resource("temp_fixtures")
        os.makedirs(fixture_dir, exist_ok=True)
        self._cache_path = os.path.join(fixture_dir, self._cache_name)
        self._pcs_runner = PcsRunner(self._cache_path)

        with open(self._empty_cib_path, "r") as template_file, open(
            self.cache_path, "w"
        ) as cache_file:
            cache_file.write(template_file.read())
        self._setup_cib()

    def clean_up(self):
        if os.path.isfile(self.cache_path):
            os.unlink(self.cache_path)

    @property
    def cache_path(self):
        if self._cache_path is None:
            raise AssertionError("Cache has not been initiialized")
        return self._cache_path

    # methods for supporting assert_pcs_success
    @property
    def pcs_runner(self):
        if self._pcs_runner is None:
            raise AssertionError("Cache has not been initialized")
        return self._pcs_runner

    def assertEqual(self, first, second, msg=None):
        # pylint: disable=invalid-name
        # pylint: disable=no-self-use
        if first != second:
            raise AssertionError(
                f"{msg}\n{first} != {second}" if msg else f"{first} != {second}"
            )


def wrap_element_by_master(cib_file, resource_id, master_id=None):
    cib_file.seek(0)
    cib_tree = etree.parse(cib_file, etree.XMLParser(huge_tree=True)).getroot()
    element = cib_tree.find(f'.//*[@id="{resource_id}"]')
    final_master_id = (
        master_id if master_id is not None else f"{resource_id}-master"
    )
    master_element = _xml_to_element(
        f"""
        <master id="{final_master_id}">
        </master>
    """
    )
    element.getparent().append(master_element)
    master_element.append(element)
    final_xml = etree_to_str(cib_tree)

    environ = dict(os.environ)
    environ["CIB_file"] = cib_file.name
    runner = CommandRunner(
        mock.MagicMock(logging.Logger), MockLibraryReportProcessor(), environ
    )
    stdout, stderr, retval = runner.run(
        [
            os.path.join(settings.pacemaker_binaries, "cibadmin"),
            "--replace",
            "--scope",
            "resources",
            "--xml-pipe",
        ],
        stdin_string=final_xml,
    )
    assert retval == 0, (
        "Error running wrap_element_by_master:\n" + stderr + "\n" + stdout
    )


def wrap_element_by_master_file(filepath, resource_id, master_id=None):
    cib_tmp = get_tmp_file("wrap_by_master")
    write_file_to_tmpfile(filepath, cib_tmp)
    wrap_element_by_master(cib_tmp, resource_id, master_id=master_id)
    cib_tmp.seek(0)
    with open(filepath, "w") as target:
        target.write(cib_tmp.read())
    cib_tmp.close()


def fixture_master_xml(name, all_ops=True, meta_dict=None):
    default_ops = f"""
            <op id="{name}-notify-interval-0s" interval="0s" name="notify"
                timeout="5"
            />
            <op id="{name}-start-interval-0s" interval="0s" name="start"
                timeout="20"
            />
            <op id="{name}-stop-interval-0s" interval="0s" name="stop"
                timeout="20"
            />
    """
    meta_xml = ""
    if meta_dict:
        meta_lines = (
            [f'<meta_attributes id="{name}-master-meta_attributes">']
            + [
                f'<nvpair id="{name}-master-meta_attributes-{key}" name="{key}" value="{val}"/>'
                for key, val in meta_dict.items()
            ]
            + ["</meta_attributes>"]
        )
        meta_xml = "\n".join(meta_lines)
    master = f"""
      <master id="{name}-master">
        <primitive class="ocf" id="{name}" provider="pacemaker" type="Stateful">
          <operations>
            <op id="{name}-monitor-interval-10" interval="10" name="monitor"
                role="Master" timeout="20"
            />
            <op id="{name}-monitor-interval-11" interval="11" name="monitor"
                role="Slave" timeout="20"
            />
    """
    if all_ops:
        master += default_ops
    master += f"""
          </operations>
        </primitive>
        {meta_xml}
      </master>
    """
    return master


def fixture_to_cib(cib_file, xml):
    environ = dict(os.environ)
    environ["CIB_file"] = cib_file
    runner = CommandRunner(
        mock.MagicMock(logging.Logger), MockLibraryReportProcessor(), environ
    )
    stdout, stderr, retval = runner.run(
        [
            os.path.join(settings.pacemaker_binaries, "cibadmin"),
            "--create",
            "--scope",
            "resources",
            "--xml-text",
            xml,
        ]
    )
    assert retval == 0, (
        "Error running fixture_to_cib:\n" + stderr + "\n" + stdout
    )


def _replace(element_to_replace, new_element):
    parent = element_to_replace.getparent()
    for child in parent:
        if element_to_replace == child:
            index = list(parent).index(child)
            parent.remove(child)
            parent.insert(index, new_element)
            return


def _xml_to_element(xml):
    try:
        new_element = etree.fromstring(xml)
    except etree.XMLSyntaxError as e:
        raise AssertionError(
            "Cannot put to the cib a non-xml fragment:\n'{0}'".format(xml)
        ) from e
    return new_element


def _find_all_in(cib_tree, element_xpath):
    element_list = cib_tree.xpath(element_xpath)
    if not element_list:
        raise AssertionError(
            "Cannot find '{0}' in given cib:\n{1}".format(
                element_xpath, etree_to_str(cib_tree)
            )
        )
    return element_list


def _find_in(cib_tree, element_xpath):
    element_list = _find_all_in(cib_tree, element_xpath)
    if len(element_list) > 1:
        raise AssertionError(
            "Found more than one '{0}' in given cib:\n{1}".format(
                element_xpath, etree_to_str(cib_tree)
            )
        )
    return element_list[0]


def remove(element_xpath):
    def _remove(cib_tree):
        xpath_list = (
            [element_xpath] if isinstance(element_xpath, str) else element_xpath
        )
        for xpath in xpath_list:
            for element_to_remove in _find_all_in(cib_tree, xpath):
                element_to_remove.getparent().remove(element_to_remove)

    return _remove


def put_or_replace(parent_xpath, new_content):
    # This tranformation makes sense in "configuration" section only. In this
    # section there are sub-tags (optional or mandatory) that can occure max 1x.
    #
    # In other sections it is possible to have more occurences of sub-tags. For
    # such cases it is better to use `replace_all` - the difference is that in
    # `replace_all` the element to be replaced is specified by full xpath
    # whilst in `put_or_replace` the xpath to the parent element is specified.
    def replace_optional(cib_tree):
        element = _xml_to_element(new_content)
        parent = _find_in(cib_tree, parent_xpath)
        current_elements = parent.findall(element.tag)

        if len(current_elements) > 1:
            raise _cannot_multireplace(element.tag, parent_xpath, cib_tree)

        if current_elements:
            _replace(current_elements[0], element)
        else:
            parent.append(element)

    return replace_optional


def replace_all(replacements):
    """
    Return a function that replace more elements (defined by replacement_dict)
    in the cib_tree with new_content.

    dict replacemens -- contains more replacements:
        key is xpath - its destination must be one element: replacement is
        applied only on the first occurence
        value is new content -contains a content that have to be placed instead
        of an element found by element_xpath
    """

    def replace(cib_tree):
        for xpath, new_content in replacements.items():
            _replace(_find_in(cib_tree, xpath), _xml_to_element(new_content))

    return replace


def append_all(append_map):
    """
    Return a function that appends more elements after specified (xpath) element
    dict append_map -- a key is an xpath pointing to a target element (for
        appending), value is appended content
    """

    def append(cib_tree):
        for xpath, new_content in append_map.items():
            _find_in(cib_tree, xpath).append(_xml_to_element(new_content))

    return append


# Possible modifier shortcuts are defined here.
# Keep in mind that every key will be named parameter in config function
# (see modifier_shortcuts param in some of pcs_test.tools.command_env.config_*
# modules)
#
# DO NOT USE CONFLICTING KEYS HERE!
# 1) args of pcs_test.tools.command_env.calls#CallListBuilder.place:
#  name, before, instead
# 2) args of pcs_test.tools.command_env.mock_runner#Call.__init__
#  command, stdout, stderr, returncode, check_stdin
# 3) special args of pcs_test.tools.command_env.config_*
#  modifiers, filename, load_key, wait, exception
# It would be not aplied. Not even mention that the majority of these names do
# not make sense for a cib modifying ;)
MODIFIER_GENERATORS = {
    "remove": remove,
    "replace": replace_all,
    "append": append_all,
    "resources": lambda xml: replace_all({"./configuration/resources": xml}),
    "constraints": lambda xml: replace_all(
        {"./configuration/constraints": xml}
    ),
    "fencing_topology": lambda xml: put_or_replace("./configuration", xml),
    "status": lambda xml: put_or_replace(".", xml),
    "tags": lambda xml: put_or_replace("./configuration", xml),
    "optional_in_conf": lambda xml: put_or_replace("./configuration", xml),
    # common modifier `put_or_replace` makes not sense - see explanation inside
    # this function - all occurences should be satisfied by `optional_in_conf`
}


def create_modifiers(**modifier_shortcuts):
    """
    Return list of modifiers: list of functions that transform cib

    dict modifier_shortcuts -- a new modifier is generated from each modifier
        shortcut.
        As key there can be keys of MODIFIER_GENERATORS.
        Value is passed into appropriate generator from MODIFIER_GENERATORS.

    """
    unknown_shortcuts = set(modifier_shortcuts.keys()) - set(
        MODIFIER_GENERATORS.keys()
    )
    if unknown_shortcuts:
        raise AssertionError(
            "Unknown modifier shortcuts '{0}', available are: '{1}'".format(
                "', '".join(list(unknown_shortcuts)),
                "', '".join(MODIFIER_GENERATORS.keys()),
            )
        )

    return [
        MODIFIER_GENERATORS[name](param)
        for name, param in modifier_shortcuts.items()
    ]


def modify_cib(cib_xml, modifiers=None, **modifier_shortcuts):
    """
    Apply modifiers to cib_xml and return the result cib_xml

    string cib_xml -- initial cib
    list of callable modifiers -- each takes cib (etree.Element)
    dict modifier_shortcuts -- a new modifier is generated from each modifier
        shortcut.
        As key there can be keys of MODIFIER_GENERATORS.
        Value is passed into appropriate generator from MODIFIER_GENERATORS.
    """
    modifiers = modifiers if modifiers else []
    all_modifiers = modifiers + create_modifiers(**modifier_shortcuts)

    if not all_modifiers:
        return cib_xml

    cib_tree = etree.fromstring(cib_xml)
    for modify in all_modifiers:
        modify(cib_tree)

    return etree_to_str(cib_tree)


def _cannot_multireplace(tag, parent_xpath, cib_tree):
    return AssertionError(
        (
            "Cannot replace '{element}' in '{parent}' because '{parent}'"
            " contains more than one '{element}' in given cib:\n{cib}"
        ).format(element=tag, parent=parent_xpath, cib=etree_to_str(cib_tree))
    )



from pcs import (
    host,
    usage,
)
from pcs.cli.common.routing import create_router


host_cmd = create_router(
    {
        "help": lambda lib, argv, modifiers: usage.host(argv),
        "auth": host.auth_cmd,
        "deauth": host.deauth_cmd,
    },
    ["host"],
)



import re

from pcs_test.tier1.cib_resource.common import ResourceTest
from pcs_test.tier1.cib_resource.stonith_common import need_load_xvm_fence_agent
from pcs_test.tools.misc import is_minimum_pacemaker_version

PCMK_2_0_3_PLUS = is_minimum_pacemaker_version(2, 0, 3)
PCMK_2_0_5_PLUS = is_minimum_pacemaker_version(2, 0, 5)
ERRORS_HAVE_OCURRED = (
    "Error: Errors have occurred, therefore pcs is unable to continue\n"
)


class PlainStonith(ResourceTest):
    @need_load_xvm_fence_agent
    def test_simplest(self):
        self.assert_effect(
            "stonith create S fence_xvm".split(),
            """<resources>
                <primitive class="stonith" id="S" type="fence_xvm">
                    <operations>
                        <op id="S-monitor-interval-60s" interval="60s"
                            name="monitor"
                        />
                    </operations>
                </primitive>
            </resources>""",
        )

    def test_base_with_agent_that_provides_unfencing(self):
        self.assert_effect(
            "stonith create S fence_scsi".split(),
            """<resources>
                <primitive class="stonith" id="S" type="fence_scsi">
                    <meta_attributes id="S-meta_attributes">
                        <nvpair id="S-meta_attributes-provides" name="provides"
                            value="unfencing"
                        />
                    </meta_attributes>
                    <operations>
                        <op id="S-monitor-interval-60s" interval="60s"
                            name="monitor"
                        />
                    </operations>
                </primitive>
            </resources>""",
        )

    def test_error_when_not_valid_name(self):
        self.assert_pcs_fail_regardless_of_force(
            "stonith create S fence_xvm:invalid".split(),
            "Error: Invalid stonith agent name 'fence_xvm:invalid'. List of"
            " agents can be obtained by using command 'pcs stonith list'."
            " Do not use the 'stonith:' prefix. Agent name cannot contain"
            " the ':' character.\n",
        )

    def test_error_when_not_valid_agent(self):
        error = error_re = None
        if PCMK_2_0_3_PLUS:
            # pacemaker 2.0.5 adds 'crm_resource:'
            # The exact message returned form pacemaker differs from version to
            # version (sometimes from commit to commit), so we don't check for
            # the whole of it.
            error_re = re.compile(
                "^"
                "Error: Agent 'absent' is not installed or does not provide "
                "valid metadata:( crm_resource:)? Metadata query for "
                "stonith:absent failed:.+"
                "use --force to override\n$",
                re.MULTILINE,
            )
        else:
            error = (
                "Error: Agent 'absent' is not installed or does not provide "
                "valid metadata: Agent absent not found or does not support "
                "meta-data: Invalid argument (22), "
                "Metadata query for stonith:absent failed: Input/output error, "
                "use --force to override\n"
            )
        self.assert_pcs_fail(
            "stonith create S absent".split(),
            stdout_full=error,
            stdout_regexp=error_re,
        )

    def test_warning_when_not_valid_agent(self):
        error = error_re = None
        if PCMK_2_0_3_PLUS:
            # pacemaker 2.0.5 adds 'crm_resource:'
            # The exact message returned form pacemaker differs from version to
            # version (sometimes from commit to commit), so we don't check for
            # the whole of it.
            error_re = re.compile(
                "^"
                "Warning: Agent 'absent' is not installed or does not provide "
                "valid metadata:( crm_resource:)? Metadata query for "
                "stonith:absent failed:.+",
                re.MULTILINE,
            )
        else:
            error = (
                "Warning: Agent 'absent' is not installed or does not provide "
                "valid metadata: Agent absent not found or does not support "
                "meta-data: Invalid argument (22), "
                "Metadata query for stonith:absent failed: Input/output error\n"
            )
        self.assert_effect(
            "stonith create S absent --force".split(),
            """<resources>
                <primitive class="stonith" id="S" type="absent">
                    <operations>
                        <op id="S-monitor-interval-60s" interval="60s"
                            name="monitor"
                        />
                    </operations>
                </primitive>
            </resources>""",
            output=error,
            output_regexp=error_re,
        )

    @need_load_xvm_fence_agent
    def test_disabled_puts_target_role_stopped(self):
        self.assert_effect(
            "stonith create S fence_xvm --disabled".split(),
            """<resources>
                <primitive class="stonith" id="S" type="fence_xvm">
                    <meta_attributes id="S-meta_attributes">
                        <nvpair id="S-meta_attributes-target-role"
                            name="target-role" value="Stopped"
                        />
                    </meta_attributes>
                    <operations>
                        <op id="S-monitor-interval-60s" interval="60s"
                            name="monitor"
                        />
                    </operations>
                </primitive>
            </resources>""",
        )

    def test_debug_and_verbose_allowed(self):
        self.assert_effect(
            "stonith create S fence_apc ip=i username=u verbose=v debug=d".split(),
            """<resources>
                <primitive class="stonith" id="S" type="fence_apc">
                    <instance_attributes id="S-instance_attributes">
                        <nvpair id="S-instance_attributes-debug"
                            name="debug" value="d"
                        />
                        <nvpair id="S-instance_attributes-ip"
                            name="ip" value="i"
                        />
                        <nvpair id="S-instance_attributes-username"
                            name="username" value="u"
                        />
                        <nvpair id="S-instance_attributes-verbose"
                            name="verbose" value="v"
                        />
                    </instance_attributes>
                    <operations>
                        <op id="S-monitor-interval-60s" interval="60s"
                            name="monitor"
                        />
                    </operations>
                </primitive>
            </resources>""",
        )

    @need_load_xvm_fence_agent
    def test_error_when_action_specified(self):
        self.assert_pcs_fail(
            "stonith create S fence_xvm action=reboot".split(),
            "Error: stonith option 'action' is deprecated and should not be"
            " used, use 'pcmk_off_action', 'pcmk_reboot_action' instead, "
            "use --force to override\n" + ERRORS_HAVE_OCURRED,
        )

    @need_load_xvm_fence_agent
    def test_warn_when_action_specified_forced(self):
        self.assert_effect(
            "stonith create S fence_xvm action=reboot --force".split(),
            """<resources>
                <primitive class="stonith" id="S" type="fence_xvm">
                    <instance_attributes id="S-instance_attributes">
                        <nvpair id="S-instance_attributes-action"
                            name="action" value="reboot"
                        />
                    </instance_attributes>
                    <operations>
                        <op id="S-monitor-interval-60s" interval="60s"
                            name="monitor"
                        />
                    </operations>
                </primitive>
            </resources>""",
            "Warning: stonith option 'action' is deprecated and should not be"
            " used, use 'pcmk_off_action', 'pcmk_reboot_action' instead\n",
        )


class WithMeta(ResourceTest):
    @need_load_xvm_fence_agent
    def test_simplest_with_meta_provides(self):
        self.assert_effect(
            "stonith create S fence_xvm meta provides=something".split(),
            """<resources>
                <primitive class="stonith" id="S" type="fence_xvm">
                    <meta_attributes id="S-meta_attributes">
                        <nvpair id="S-meta_attributes-provides" name="provides"
                            value="something"
                        />
                    </meta_attributes>
                    <operations>
                        <op id="S-monitor-interval-60s" interval="60s"
                            name="monitor"
                        />
                    </operations>
                </primitive>
            </resources>""",
        )

    def test_base_with_agent_that_provides_unfencing_with_meta_provides(self):
        self.assert_effect(
            "stonith create S fence_scsi meta provides=something".split(),
            """<resources>
                <primitive class="stonith" id="S" type="fence_scsi">
                    <meta_attributes id="S-meta_attributes">
                        <nvpair id="S-meta_attributes-provides" name="provides"
                            value="unfencing"
                        />
                    </meta_attributes>
                    <operations>
                        <op id="S-monitor-interval-60s" interval="60s"
                            name="monitor"
                        />
                    </operations>
                </primitive>
            </resources>""",
        )


class InGroup(ResourceTest):
    @need_load_xvm_fence_agent
    def test_command_simply_puts_stonith_into_group(self):
        self.assert_effect(
            "stonith create S fence_xvm --group G".split(),
            """<resources>
                <group id="G">
                    <primitive class="stonith" id="S" type="fence_xvm">
                        <operations>
                            <op id="S-monitor-interval-60s" interval="60s"
                                name="monitor"
                            />
                        </operations>
                    </primitive>
                </group>
            </resources>""",
        )

    @need_load_xvm_fence_agent
    def test_command_simply_puts_stonith_into_group_at_the_end(self):
        self.assert_pcs_success("stonith create S1 fence_xvm --group G".split())
        self.assert_effect(
            "stonith create S2 fence_xvm --group G".split(),
            """<resources>
                <group id="G">
                    <primitive class="stonith" id="S1" type="fence_xvm">
                        <operations>
                            <op id="S1-monitor-interval-60s" interval="60s"
                                name="monitor"
                            />
                        </operations>
                    </primitive>
                    <primitive class="stonith" id="S2" type="fence_xvm">
                        <operations>
                            <op id="S2-monitor-interval-60s" interval="60s"
                                name="monitor"
                            />
                        </operations>
                    </primitive>
                </group>
            </resources>""",
        )

    @need_load_xvm_fence_agent
    def test_command_simply_puts_stonith_into_group_before_another(self):
        self.assert_pcs_success("stonith create S1 fence_xvm --group G".split())
        self.assert_effect(
            "stonith create S2 fence_xvm --group G --before S1".split(),
            """<resources>
                <group id="G">
                    <primitive class="stonith" id="S2" type="fence_xvm">
                        <operations>
                            <op id="S2-monitor-interval-60s" interval="60s"
                                name="monitor"
                            />
                        </operations>
                    </primitive>
                    <primitive class="stonith" id="S1" type="fence_xvm">
                        <operations>
                            <op id="S1-monitor-interval-60s" interval="60s"
                                name="monitor"
                            />
                        </operations>
                    </primitive>
                </group>
            </resources>""",
        )

    @need_load_xvm_fence_agent
    def test_command_simply_puts_stonith_into_group_after_another(self):
        self.assert_pcs_success_all(
            [
                "stonith create S1 fence_xvm --group G".split(),
                "stonith create S2 fence_xvm --group G".split(),
            ]
        )
        self.assert_effect(
            "stonith create S3 fence_xvm --group G --after S1".split(),
            """<resources>
                <group id="G">
                    <primitive class="stonith" id="S1" type="fence_xvm">
                        <operations>
                            <op id="S1-monitor-interval-60s" interval="60s"
                                name="monitor"
                            />
                        </operations>
                    </primitive>
                    <primitive class="stonith" id="S3" type="fence_xvm">
                        <operations>
                            <op id="S3-monitor-interval-60s" interval="60s"
                                name="monitor"
                            />
                        </operations>
                    </primitive>
                    <primitive class="stonith" id="S2" type="fence_xvm">
                        <operations>
                            <op id="S2-monitor-interval-60s" interval="60s"
                                name="monitor"
                            />
                        </operations>
                    </primitive>
                </group>
            </resources>""",
        )

    @need_load_xvm_fence_agent
    def test_fail_when_inteded_before_item_does_not_exist(self):
        self.assert_pcs_fail(
            "stonith create S2 fence_xvm --group G --before S1".split(),
            "Error: there is no resource 'S1' in the group 'G'\n",
        )

    @need_load_xvm_fence_agent
    def test_fail_when_inteded_after_item_does_not_exist(self):
        self.assert_pcs_fail(
            "stonith create S2 fence_xvm --group G --after S1".split(),
            "Error: there is no resource 'S1' in the group 'G'\n",
        )

    def test_fail_when_entered_both_after_and_before(self):
        self.assert_pcs_fail(
            "stonith create S fence_xvm --group G --after S1 --before S2".split(),
            "Error: you cannot specify both --before and --after\n",
        )

    def test_fail_when_after_is_used_without_group(self):
        self.assert_pcs_fail(
            "stonith create S fence_xvm --after S1".split(),
            "Error: you cannot use --after without --group\n",
        )

    def test_fail_when_before_is_used_without_group(self):
        self.assert_pcs_fail(
            "stonith create S fence_xvm --before S1".split(),
            "Error: you cannot use --before without --group\n",
        )

    def test_fail_when_before_after_conflicts_and_moreover_without_group(self):
        self.assert_pcs_fail(
            "stonith create S fence_xvm --after S1 --before S2".split(),
            "Error: you cannot specify both --before and --after"
            " and you have to specify --group\n",
        )



from typing import Iterable, Mapping, Optional

from lxml import etree

from pcs.common import reports
from pcs.common.reports import (
    codes as report_codes,
    ReportProcessor,
)
from pcs.common.reports.item import ReportItem
from pcs.lib.cib.nvpair import (
    append_new_instance_attributes,
    append_new_meta_attributes,
    get_value,
    get_nvset_as_dict,
)
from pcs.lib.cib.resource.operations import (
    prepare as prepare_operations,
    create_operations,
)
from pcs.lib.cib.tools import (
    does_id_exist,
    find_element_by_tag_and_id,
    are_new_role_names_supported,
)
from pcs.lib.errors import LibraryError
from pcs.lib.pacemaker.values import validate_id


TAG = "primitive"


def is_primitive(resource_el):
    return resource_el.tag == TAG


def find_primitives_by_agent(resources_section, resource_agent_obj):
    """
    Returns list of primitive resource elements which are using same resource
    agent as specified by resource_agent_obj.

    resources_section etree.Element -- element <resources/> from CIB
    resource_agent_obj pcs.lib.resource_agent.CrmAgent -- agent of which
        resources should be returned
    """
    provider = resource_agent_obj.get_provider()
    return resources_section.xpath(
        ".//primitive[@class='{_class}' and @type='{_type}'{_provider}]".format(
            _class=resource_agent_obj.get_standard(),
            _type=resource_agent_obj.get_type(),
            _provider=f" and @provider='{provider}'" if provider else "",
        )
    )


def create(
    report_processor: ReportProcessor,
    resources_section,
    id_provider,
    resource_id,
    resource_agent,
    raw_operation_list: Optional[Iterable[Mapping[str, str]]] = None,
    meta_attributes=None,
    instance_attributes=None,
    allow_invalid_operation=False,
    allow_invalid_instance_attributes=False,
    use_default_operations=True,
    resource_type="resource",
    # TODO remove this arg
    do_not_report_instance_attribute_server_exists=False,
):
    # pylint: disable=too-many-arguments
    """
    Prepare all parts of primitive resource and append it into cib.

    report_processor is a tool for warning/info/error reporting
    etree.Element resources_section is place where new element will be appended
    IdProvider id_provider -- elements' ids generator
    string resource_id is id of new resource
    lib.resource_agent.CrmAgent resource_agent
    list of dict raw_operation_list specifies operations of resource
    dict meta_attributes specifies meta attributes of resource
    dict instance_attributes specifies instance attributes of resource
    bool allow_invalid_operation is flag for skipping validation of operations
    bool allow_invalid_instance_attributes is flag for skipping validation of
        instance_attributes
    bool use_default_operations is flag for completion operations with default
        actions specified in resource agent
    string resource_type -- describes the resource for reports
    bool do_not_report_instance_attribute_server_exists -- dirty fix due to
        suboptimal architecture, TODO: fix the architecture and remove the param
    """
    if raw_operation_list is None:
        raw_operation_list = []
    if meta_attributes is None:
        meta_attributes = {}
    if instance_attributes is None:
        instance_attributes = {}

    if does_id_exist(resources_section, resource_id):
        raise LibraryError(
            ReportItem.error(reports.messages.IdAlreadyExists(resource_id))
        )
    validate_id(resource_id, "{0} name".format(resource_type))

    operation_list = prepare_operations(
        report_processor,
        raw_operation_list,
        resource_agent.get_cib_default_actions(
            necessary_only=not use_default_operations
        ),
        [operation.name for operation in resource_agent.get_actions()],
        are_new_role_names_supported(resources_section),
        allow_invalid=allow_invalid_operation,
    )

    if report_processor.report_list(
        validate_resource_instance_attributes_create(
            resource_agent,
            instance_attributes,
            resources_section,
            force=allow_invalid_instance_attributes,
            do_not_report_instance_attribute_server_exists=(
                do_not_report_instance_attribute_server_exists
            ),
        )
    ).has_errors:
        raise LibraryError()

    return append_new(
        resources_section,
        id_provider,
        resource_id,
        resource_agent.get_standard(),
        resource_agent.get_provider(),
        resource_agent.get_type(),
        instance_attributes=instance_attributes,
        meta_attributes=meta_attributes,
        operation_list=operation_list,
    )


def append_new(
    resources_section,
    id_provider,
    resource_id,
    standard,
    provider,
    agent_type,
    instance_attributes=None,
    meta_attributes=None,
    operation_list=None,
):
    # pylint:disable=too-many-arguments
    """
    Append a new primitive element to the resources_section.

    etree.Element resources_section is place where new element will be appended
    IdProvider id_provider -- elements' ids generator
    string resource_id is id of new resource
    string standard is a standard of resource agent (e.g. ocf)
    string agent_type is a type of resource agent (e.g. IPaddr2)
    string provider is a provider of resource agent (e.g. heartbeat)
    dict instance_attributes will be nvpairs inside instance_attributes element
    dict meta_attributes will be nvpairs inside meta_attributes element
    list operation_list contains dicts representing operations
        (e.g. [{"name": "monitor"}, {"name": "start"}])
    """
    attributes = {
        "id": resource_id,
        "class": standard,
        "type": agent_type,
    }
    if provider:
        attributes["provider"] = provider
    primitive_element = etree.SubElement(resources_section, TAG, attributes)

    if instance_attributes:
        append_new_instance_attributes(
            primitive_element, instance_attributes, id_provider
        )

    if meta_attributes:
        append_new_meta_attributes(
            primitive_element, meta_attributes, id_provider
        )

    create_operations(
        primitive_element, id_provider, operation_list if operation_list else []
    )

    return primitive_element


def validate_unique_instance_attributes(
    resource_agent,
    instance_attributes,
    resources_section,
    resource_id=None,
    force=False,
):
    report_list = []
    ra_unique_attributes = [
        param.name
        for param in resource_agent.get_full_info().parameters
        if param.unique
    ]
    same_agent_resources = find_primitives_by_agent(
        resources_section, resource_agent
    )
    for attr in ra_unique_attributes:
        if attr not in instance_attributes:
            continue
        conflicting_resources = {
            primitive.get("id")
            for primitive in same_agent_resources
            if (
                primitive.get("id") != resource_id
                and instance_attributes[attr]
                == get_value("instance_attributes", primitive, attr)
            )
        }
        if conflicting_resources:
            report_list.append(
                ReportItem(
                    severity=reports.item.get_severity(
                        report_codes.FORCE,
                        force,
                    ),
                    message=reports.messages.ResourceInstanceAttrValueNotUnique(
                        attr,
                        instance_attributes[attr],
                        resource_agent.get_name_info().name,
                        sorted(conflicting_resources),
                    ),
                )
            )
    return report_list


def validate_resource_instance_attributes_create(
    resource_agent,
    instance_attributes,
    resources_section,
    force=False,
    do_not_report_instance_attribute_server_exists=False,
):
    return resource_agent.validate_parameters_create(
        instance_attributes,
        force=force,
        do_not_report_instance_attribute_server_exists=(
            do_not_report_instance_attribute_server_exists
        ),
    ) + validate_unique_instance_attributes(
        resource_agent, instance_attributes, resources_section, force=force
    )


def validate_resource_instance_attributes_update(
    resource_agent,
    instance_attributes,
    resource_id,
    resources_section,
    force=False,
):
    return resource_agent.validate_parameters_update(
        get_nvset_as_dict(
            "instance_attributes",
            find_element_by_tag_and_id(
                "primitive", resources_section, resource_id
            ),
        ),
        instance_attributes,
        force=force,
    ) + validate_unique_instance_attributes(
        resource_agent,
        instance_attributes,
        resources_section,
        resource_id=resource_id,
        force=force,
    )



import re
from unittest import TestCase
from lxml import etree

from pcs_test.tools import fixture
from pcs_test.tools.assertions import assert_report_item_list_equal

from pcs.common import reports
from pcs.lib import validate
from pcs.lib.cib.tools import IdProvider

# pylint: disable=no-self-use
# pylint: disable=too-many-lines

### normalization


class ValuesToPairs(TestCase):
    def test_create_from_plain_values(self):
        self.assertEqual(
            {
                "first": validate.ValuePair("A", "a"),
                "second": validate.ValuePair("B", "b"),
            },
            validate.values_to_pairs(
                {
                    "first": "A",
                    "second": "B",
                },
                lambda key, value: value.lower(),
            ),
        )

    def test_keep_pair_if_is_already_there(self):
        self.assertEqual(
            {
                "first": validate.ValuePair("A", "aaa"),
                "second": validate.ValuePair("B", "b"),
            },
            validate.values_to_pairs(
                {
                    "first": validate.ValuePair("A", "aaa"),
                    "second": "B",
                },
                lambda key, value: value.lower(),
            ),
        )


class PairsToValues(TestCase):
    def test_keep_values_if_is_not_pair(self):
        self.assertEqual(
            {
                "first": "A",
                "second": "B",
            },
            validate.pairs_to_values(
                {
                    "first": "A",
                    "second": "B",
                }
            ),
        )

    def test_extract_normalized_values(self):
        self.assertEqual(
            {
                "first": "aaa",
                "second": "B",
            },
            validate.pairs_to_values(
                {
                    "first": validate.ValuePair(original="A", normalized="aaa"),
                    "second": "B",
                }
            ),
        )


class OptionValueNormalization(TestCase):
    def test_return_normalized_value_if_normalization_for_key_specified(self):
        normalize = validate.option_value_normalization(
            {"first": lambda value: value.upper()}
        )
        self.assertEqual("ONE", normalize("first", "one"))

    def test_return_value_if_normalization_for_key_unspecified(self):
        normalize = validate.option_value_normalization({})
        self.assertEqual("one", normalize("first", "one"))


### compound validators


class ValidatorAll(TestCase):
    def test_collect_all_errors_from_specifications(self):
        assert_report_item_list_equal(
            validate.ValidatorAll(
                [
                    validate.NamesIn(["x", "y"]),
                    validate.MutuallyExclusive(["x", "y"]),
                    validate.ValuePositiveInteger("x"),
                    validate.ValueIn("y", ["a", "b"]),
                ]
            ).validate(
                {
                    "x": "abcd",
                    "y": "defg",
                    "z": "hijk",
                }
            ),
            [
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    option_names=["z"],
                    option_type=None,
                    allowed=["x", "y"],
                    allowed_patterns=[],
                ),
                fixture.error(
                    reports.codes.MUTUALLY_EXCLUSIVE_OPTIONS,
                    option_names=["x", "y"],
                    option_type=None,
                ),
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_value="abcd",
                    option_name="x",
                    allowed_values="a positive integer",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_value="defg",
                    option_name="y",
                    allowed_values=["a", "b"],
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
        )


class ValidatorFirstError(TestCase):
    class Validator(validate.ValueValidator):
        def _validate_value(self, value):
            severity = None
            if value.normalized == "a":
                severity = reports.ReportItemSeverity.error()
            if value.normalized == "b":
                severity = reports.ReportItemSeverity.warning()
            if severity is None:
                return []
            return [
                reports.item.ReportItem(
                    severity,
                    reports.messages.InvalidOptionValue(
                        self._option_name,
                        value.original,
                        "test report",
                    ),
                )
            ]

    def setUp(self):
        self.validator = validate.ValidatorFirstError(
            [
                self.Validator("name1"),
                self.Validator("name2"),
            ]
        )

    def test_no_reports(self):
        assert_report_item_list_equal(
            self.validator.validate({"name1": "c", "name2": "d"}), []
        )

    def test_first_errors(self):
        assert_report_item_list_equal(
            self.validator.validate({"name1": "a", "name2": "d"}),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="name1",
                    option_value="a",
                    allowed_values="test report",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
        )

    def test_second_errors(self):
        assert_report_item_list_equal(
            self.validator.validate({"name1": "c", "name2": "a"}),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="name2",
                    option_value="a",
                    allowed_values="test report",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
        )

    def test_keep_warnings(self):
        assert_report_item_list_equal(
            self.validator.validate({"name1": "b", "name2": "a"}),
            [
                fixture.warn(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="name1",
                    option_value="b",
                    allowed_values="test report",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="name2",
                    option_value="a",
                    allowed_values="test report",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
        )


### keys validators


class CorosyncOption(TestCase):
    def test_valid(self):
        assert_report_item_list_equal(
            validate.CorosyncOption().validate(
                {
                    "name_-/NAME09": "value",
                }
            ),
            [],
        )

    def test_forbidden_characters(self):
        bad_names = [f"na{char}me" for char in ".: {}#č"]
        assert_report_item_list_equal(
            validate.CorosyncOption().validate(
                {name: "value" for name in bad_names}
            ),
            [
                fixture.error(
                    reports.codes.INVALID_USERDEFINED_OPTIONS,
                    option_names=sorted(bad_names),
                    option_type=None,
                    allowed_characters="a-z A-Z 0-9 /_-",
                )
            ],
        )

    def test_option_type(self):
        bad_names = [f"na{char}me" for char in ".: {}#č"]
        assert_report_item_list_equal(
            validate.CorosyncOption(option_type="type").validate(
                {name: "value" for name in bad_names}
            ),
            [
                fixture.error(
                    reports.codes.INVALID_USERDEFINED_OPTIONS,
                    option_names=sorted(bad_names),
                    option_type="type",
                    allowed_characters="a-z A-Z 0-9 /_-",
                )
            ],
        )


class DependsOnOption(TestCase):
    def test_success_when_dependency_present(self):
        assert_report_item_list_equal(
            validate.DependsOnOption(
                ["name"], "prerequisite", option_type="type"
            ).validate(
                {
                    "name": "value",
                    "prerequisite": "value",
                }
            ),
            [],
        )

    def test_report_when_dependency_missing(self):
        assert_report_item_list_equal(
            validate.DependsOnOption(
                ["name"],
                "prerequisite",
                option_type="type1",
                prerequisite_type="type2",
            ).validate({"name": "value"}),
            [
                fixture.error(
                    reports.codes.PREREQUISITE_OPTION_IS_MISSING,
                    option_name="name",
                    option_type="type1",
                    prerequisite_name="prerequisite",
                    prerequisite_type="type2",
                ),
            ],
        )

    def test_more_options(self):
        assert_report_item_list_equal(
            validate.DependsOnOption(
                ["name1", "name2", "name3"], "prerequisite"
            ).validate(
                {
                    "name1": "value",
                    "name3": "value",
                }
            ),
            [
                fixture.error(
                    reports.codes.PREREQUISITE_OPTION_IS_MISSING,
                    option_name=name,
                    option_type=None,
                    prerequisite_name="prerequisite",
                    prerequisite_type=None,
                )
                for name in ["name1", "name3"]
            ],
        )


class IsRequiredAll(TestCase):
    def test_returns_no_report_when_required_is_present(self):
        assert_report_item_list_equal(
            validate.IsRequiredAll(["name"], "some type").validate(
                {"name": "monitor"}
            ),
            [],
        )

    def test_returns_report_when_required_is_missing(self):
        assert_report_item_list_equal(
            validate.IsRequiredAll(["name"], "some type").validate({}),
            [
                fixture.error(
                    reports.codes.REQUIRED_OPTIONS_ARE_MISSING,
                    option_names=["name"],
                    option_type="some type",
                ),
            ],
        )

    def test_more_options(self):
        assert_report_item_list_equal(
            validate.IsRequiredAll(
                ["name1", "name2", "name3"], "some type"
            ).validate({"name2": "value2"}),
            [
                fixture.error(
                    reports.codes.REQUIRED_OPTIONS_ARE_MISSING,
                    option_names=["name1", "name3"],
                    option_type="some type",
                ),
            ],
        )


class IsRequiredSome(TestCase):
    def test_returns_no_report_when_first_is_present(self):
        assert_report_item_list_equal(
            validate.IsRequiredSome(["first", "second"], "type").validate(
                {"first": "value"}
            ),
            [],
        )

    def test_returns_no_report_when_second_is_present(self):
        assert_report_item_list_equal(
            validate.IsRequiredSome(["first", "second"], "type").validate(
                {"second": "value"}
            ),
            [],
        )

    def test_returns_report_when_missing(self):
        assert_report_item_list_equal(
            validate.IsRequiredSome(["first", "second"], "type").validate(
                {"third": "value"}
            ),
            [
                fixture.error(
                    reports.codes.REQUIRED_OPTION_OF_ALTERNATIVES_IS_MISSING,
                    option_names=["first", "second"],
                    option_type="type",
                ),
            ],
        )


class MutuallyExclusive(TestCase):
    def test_returns_empty_report_when_valid(self):
        assert_report_item_list_equal(
            validate.MutuallyExclusive(["a", "b"]).validate({"a": "A"}),
            [],
        )

    def test_returns_mutually_exclusive_report_on_2_names_conflict(self):
        assert_report_item_list_equal(
            validate.MutuallyExclusive(["a", "b", "c"]).validate(
                {
                    "a": "A",
                    "b": "B",
                    "d": "D",
                }
            ),
            [
                fixture.error(
                    reports.codes.MUTUALLY_EXCLUSIVE_OPTIONS,
                    option_type=None,
                    option_names=["a", "b"],
                ),
            ],
        )

    def test_returns_mutually_exclusive_report_on_multiple_name_conflict(self):
        assert_report_item_list_equal(
            validate.MutuallyExclusive(
                ["a", "b", "c", "e"], option_type="option"
            ).validate(
                {
                    "a": "A",
                    "b": "B",
                    "c": "C",
                    "d": "D",
                }
            ),
            [
                fixture.error(
                    reports.codes.MUTUALLY_EXCLUSIVE_OPTIONS,
                    option_type="option",
                    option_names=["a", "b", "c"],
                ),
            ],
        )


class NamesIn(TestCase):
    def test_return_empty_report_on_allowed_names(self):
        assert_report_item_list_equal(
            validate.NamesIn(["a", "b", "c"]).validate({"a": "A", "b": "B"}), []
        )

    def test_return_error_on_not_allowed_names(self):
        assert_report_item_list_equal(
            validate.NamesIn(["a", "b", "c"], option_type="option").validate(
                {"x": "X", "y": "Y"}
            ),
            [
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    option_names=["x", "y"],
                    allowed=["a", "b", "c"],
                    option_type="option",
                    allowed_patterns=[],
                )
            ],
        )

    def test_return_error_on_banned_names(self):
        assert_report_item_list_equal(
            validate.NamesIn(
                ["a", "b"], banned_name_list=["x", "y", "z"]
            ).validate({"x": "X", "a": "A", "z": "Z"}),
            [
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    option_names=["x", "z"],
                    allowed=["a", "b"],
                    option_type=None,
                    allowed_patterns=[],
                )
            ],
        )

    def test_return_error_on_not_allowed_and_banned_names(self):
        assert_report_item_list_equal(
            validate.NamesIn(
                ["a", "b"], banned_name_list=["x", "y", "z"]
            ).validate({"x": "X", "a": "A", "z": "Z", "c": "C"}),
            [
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    option_names=["c", "x", "z"],
                    allowed=["a", "b"],
                    option_type=None,
                    allowed_patterns=[],
                )
            ],
        )

    def test_return_error_on_not_allowed_and_banned_names_forceable(self):
        code = "force_code"
        assert_report_item_list_equal(
            validate.NamesIn(
                ["a", "b"],
                banned_name_list=["x", "y", "z"],
                severity=reports.item.ReportItemSeverity.error(code),
            ).validate({"x": "X", "a": "A", "z": "Z", "c": "C", "d": "D"}),
            [
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    force_code=code,
                    option_names=["c", "d"],
                    allowed=["a", "b"],
                    option_type=None,
                    allowed_patterns=[],
                ),
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    option_names=["x", "z"],
                    allowed=["a", "b"],
                    option_type=None,
                    allowed_patterns=[],
                ),
            ],
        )

    def test_return_error_on_not_allowed_and_banned_names_forced(self):
        assert_report_item_list_equal(
            validate.NamesIn(
                ["a", "b"],
                banned_name_list=["x", "y", "z"],
                severity=reports.item.ReportItemSeverity.warning(),
            ).validate({"x": "X", "a": "A", "z": "Z", "c": "C", "d": "D"}),
            [
                fixture.warn(
                    reports.codes.INVALID_OPTIONS,
                    option_names=["c", "d"],
                    allowed=["a", "b"],
                    option_type=None,
                    allowed_patterns=[],
                ),
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    option_names=["x", "z"],
                    allowed=["a", "b"],
                    option_type=None,
                    allowed_patterns=[],
                ),
            ],
        )

    def test_return_error_with_allowed_patterns(self):
        assert_report_item_list_equal(
            validate.NamesIn(
                ["a", "b", "c"], allowed_option_patterns=["pattern"]
            ).validate({"x": "X", "y": "Y"}),
            [
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    option_names=["x", "y"],
                    allowed=["a", "b", "c"],
                    option_type=None,
                    allowed_patterns=["pattern"],
                )
            ],
        )

    def test_return_forceable_error_on_not_allowed_names(self):
        assert_report_item_list_equal(
            validate.NamesIn(
                ["a", "b", "c"],
                option_type="some option",
                severity=reports.item.ReportItemSeverity.error("FORCE_CODE"),
            ).validate({"x": "X", "y": "Y"}),
            [
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    force_code="FORCE_CODE",
                    option_names=["x", "y"],
                    allowed=["a", "b", "c"],
                    option_type="some option",
                    allowed_patterns=[],
                )
            ],
        )

    def test_return_warning_on_not_allowed_names(self):
        assert_report_item_list_equal(
            validate.NamesIn(
                ["a", "b", "c"],
                option_type="some option",
                severity=reports.item.ReportItemSeverity.warning(),
            ).validate({"x": "X", "y": "Y"}),
            [
                fixture.warn(
                    reports.codes.INVALID_OPTIONS,
                    option_names=["x", "y"],
                    allowed=["a", "b", "c"],
                    option_type="some option",
                    allowed_patterns=[],
                )
            ],
        )


### values validators


class ValueValidatorImplementation(validate.ValueValidator):
    def _validate_value(self, value):
        return [
            reports.item.ReportItem.error(
                reports.messages.InvalidOptionValue(
                    self._option_name,
                    value.original,
                    "test report",
                )
            )
        ]


class ValueValidator(TestCase):
    def test_value_not_specified(self):
        assert_report_item_list_equal(
            ValueValidatorImplementation("name").validate({"name1": "value1"}),
            [],
        )

    def test_value_empty_and_empty_not_allowed(self):
        assert_report_item_list_equal(
            ValueValidatorImplementation("name").validate({"name": ""}),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="name",
                    option_value="",
                    allowed_values="test report",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                )
            ],
        )

    def test_value_empty_and_empty_allowed(self):
        validator = ValueValidatorImplementation("name")
        validator.empty_string_valid = True
        assert_report_item_list_equal(validator.validate({"name": ""}), [])

    def test_value_not_valid(self):
        assert_report_item_list_equal(
            ValueValidatorImplementation("name").validate({"name": "value"}),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="name",
                    option_value="value",
                    allowed_values="test report",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                )
            ],
        )


class ValuePredicateImplementation(validate.ValuePredicateBase):
    def _is_valid(self, value):
        return value == "b"

    def _get_allowed_values(self):
        return "allowed values"

    def set_value_cannot_be_empty(self, value):
        self._value_cannot_be_empty = value
        return self

    def set_forbidden_characters(self, value):
        self._forbidden_characters = value
        return self


class ValuePredicateBase(TestCase):
    def test_returns_empty_report_on_valid_option(self):
        assert_report_item_list_equal(
            ValuePredicateImplementation("a").validate({"a": "b"}), []
        )

    def test_returns_empty_report_on_valid_normalized_option(self):
        assert_report_item_list_equal(
            ValuePredicateImplementation("a").validate(
                {"a": validate.ValuePair(original="C", normalized="b")}
            ),
            [],
        )

    def test_returns_report_about_invalid_option(self):
        assert_report_item_list_equal(
            ValuePredicateImplementation("a").validate({"a": "c"}),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="a",
                    option_value="c",
                    allowed_values="allowed values",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                )
            ],
        )

    def test_value_cannot_be_empty(self):
        assert_report_item_list_equal(
            ValuePredicateImplementation("a")
            .set_value_cannot_be_empty(True)
            .validate({"a": "c"}),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="a",
                    option_value="c",
                    allowed_values="allowed values",
                    cannot_be_empty=True,
                    forbidden_characters=None,
                )
            ],
        )

    def test_forbidden_characters(self):
        assert_report_item_list_equal(
            ValuePredicateImplementation("a")
            .set_forbidden_characters("xyz")
            .validate({"a": "c"}),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="a",
                    option_value="c",
                    allowed_values="allowed values",
                    cannot_be_empty=False,
                    forbidden_characters="xyz",
                )
            ],
        )

    def test_support_option_value_pair(self):
        assert_report_item_list_equal(
            ValuePredicateImplementation("a").validate(
                {"a": validate.ValuePair(original="b", normalized="c")}
            ),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="a",
                    option_value="b",
                    allowed_values="allowed values",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                )
            ],
        )

    def test_supports_another_report_option_name(self):
        assert_report_item_list_equal(
            ValuePredicateImplementation(
                "a", option_name_for_report="option a"
            ).validate({"a": "c"}),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="option a",
                    option_value="c",
                    allowed_values="allowed values",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                )
            ],
        )

    def test_supports_forceable_errors(self):
        assert_report_item_list_equal(
            ValuePredicateImplementation(
                "a", severity=reports.item.ReportItemSeverity.error("FORCE")
            ).validate({"a": "c"}),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    force_code="FORCE",
                    option_name="a",
                    option_value="c",
                    allowed_values="allowed values",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                )
            ],
        )

    def test_supports_warning(self):
        assert_report_item_list_equal(
            ValuePredicateImplementation(
                "a", severity=reports.item.ReportItemSeverity.warning()
            ).validate({"a": "c"}),
            [
                fixture.warn(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="a",
                    option_value="c",
                    allowed_values="allowed values",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                )
            ],
        )


class ValueCorosyncValue(TestCase):
    def test_value_ok(self):
        assert_report_item_list_equal(
            validate.ValueCorosyncValue("a").validate({"a": "valid_value"}), []
        )

    def test_empty_value(self):
        assert_report_item_list_equal(
            validate.ValueCorosyncValue("a").validate({"a": ""}), []
        )

    def test_escaped_new_lines(self):
        assert_report_item_list_equal(
            validate.ValueCorosyncValue("a").validate({"a": "\\n\\r"}), []
        )

    def test_forbidden_characters_reported(self):
        bad_value_list = [
            "{",
            "}",
            "\n",
            "\r",
            "bad{value",
            "bad}value",
            "bad\nvalue",
            "bad\rvalue",
            "value\r\nsection {\n\rnew_key: new_value\r\n}\n\r",
        ]
        for value in bad_value_list:
            with self.subTest(value=value):
                assert_report_item_list_equal(
                    validate.ValueCorosyncValue("a").validate({"a": value}),
                    [
                        fixture.error(
                            reports.codes.INVALID_OPTION_VALUE,
                            option_value=value,
                            option_name="a",
                            allowed_values=None,
                            cannot_be_empty=False,
                            forbidden_characters=r"{}\n\r",
                        ),
                    ],
                )


class ValueFloat(TestCase):
    # The real code only calls ValuePredicateBase and is_float which are both
    # heavily tested on their own => only basic tests here.
    def fixture_validator(self):
        return validate.ValueFloat("key")

    def test_empty_report_on_valid_option(self):
        assert_report_item_list_equal(
            self.fixture_validator().validate({"key": "2.3"}), []
        )

    def test_report_invalid_value(self):
        assert_report_item_list_equal(
            self.fixture_validator().validate({"key": "6a"}),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="key",
                    option_value="6a",
                    allowed_values="a floating-point number",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
        )


class ValueId(TestCase):
    def test_empty_id(self):
        assert_report_item_list_equal(
            validate.ValueId("id").validate({"id": ""}),
            [
                fixture.error(
                    reports.codes.INVALID_ID_IS_EMPTY,
                    id_description="id",
                ),
            ],
        )

    def test_invalid_first_char(self):
        assert_report_item_list_equal(
            validate.ValueId("id", option_name_for_report="test id").validate(
                {"id": "0-test"}
            ),
            [
                fixture.error(
                    reports.codes.INVALID_ID_BAD_CHAR,
                    id="0-test",
                    id_description="test id",
                    invalid_character="0",
                    is_first_char=True,
                ),
            ],
        )

    def test_invalid_char(self):
        assert_report_item_list_equal(
            validate.ValueId("id").validate({"id": "te#st"}),
            [
                fixture.error(
                    reports.codes.INVALID_ID_BAD_CHAR,
                    id="te#st",
                    id_description="id",
                    invalid_character="#",
                    is_first_char=False,
                ),
            ],
        )

    def test_used_id(self):
        id_provider = IdProvider(etree.fromstring("<a><test id='used' /></a>"))
        assert_report_item_list_equal(
            validate.ValueId("id", id_provider=id_provider).validate(
                {"id": "used"}
            ),
            [
                fixture.error(
                    reports.codes.ID_ALREADY_EXISTS,
                    id="used",
                ),
            ],
        )

    def test_pair_invalid(self):
        assert_report_item_list_equal(
            validate.ValueId("id").validate(
                {"id": validate.ValuePair("@&#", "")}
            ),
            [
                fixture.error(
                    reports.codes.INVALID_ID_IS_EMPTY,
                    # TODO: This should be INVALID_ID_BAD_CHAR with value
                    # "@&#". However an old validator is used and it doesn't
                    # work with pairs and therefore the empty string is used.
                    id_description="id",
                ),
            ],
        )

    def test_pair_used_id(self):
        id_provider = IdProvider(etree.fromstring("<a><test id='used' /></a>"))
        assert_report_item_list_equal(
            validate.ValueId("id", id_provider=id_provider).validate(
                {"id": validate.ValuePair("not-used", "used")}
            ),
            [
                fixture.error(
                    reports.codes.ID_ALREADY_EXISTS,
                    # TODO: This should be "not-used". However an old
                    # validator is used and it doesn't work with pairs.
                    id="used",
                ),
            ],
        )

    def test_success(self):
        id_provider = IdProvider(etree.fromstring("<a><test id='used' /></a>"))
        assert_report_item_list_equal(
            validate.ValueId("id", id_provider=id_provider).validate(
                {"id": "correct"}
            ),
            [],
        )

    def test_pair_success(self):
        id_provider = IdProvider(etree.fromstring("<a><test id='used' /></a>"))
        assert_report_item_list_equal(
            validate.ValueId("id", id_provider=id_provider).validate(
                {"id": validate.ValuePair("correct", "correct")}
            ),
            [],
        )


class ValueIn(TestCase):
    def test_returns_empty_report_on_valid_option(self):
        assert_report_item_list_equal(
            validate.ValueIn("a", ["b"]).validate({"a": "b"}), []
        )

    def test_returns_empty_report_on_valid_normalized_option(self):
        assert_report_item_list_equal(
            validate.ValueIn("a", ["b"]).validate(
                {"a": validate.ValuePair(original="C", normalized="b")}
            ),
            [],
        )

    def test_returns_report_about_invalid_option(self):
        assert_report_item_list_equal(
            validate.ValueIn("a", ["b"]).validate({"a": "c"}),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="a",
                    option_value="c",
                    allowed_values=["b"],
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
        )

    def test_support_option_value_pair(self):
        assert_report_item_list_equal(
            validate.ValueIn("a", ["b"]).validate(
                {"a": validate.ValuePair(original="C", normalized="c")}
            ),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="a",
                    option_value="C",
                    allowed_values=["b"],
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
        )

    def test_supports_another_report_option_name(self):
        assert_report_item_list_equal(
            validate.ValueIn(
                "a", ["b"], option_name_for_report="option a"
            ).validate({"a": "c"}),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="option a",
                    option_value="c",
                    allowed_values=["b"],
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
        )

    def test_supports_forceable_errors(self):
        assert_report_item_list_equal(
            validate.ValueIn(
                "a",
                ["b"],
                severity=reports.item.ReportItemSeverity.error("FORCE"),
            ).validate({"a": "c"}),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    force_code="FORCE",
                    option_name="a",
                    option_value="c",
                    allowed_values=["b"],
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
        )

    def test_supports_warning(self):
        assert_report_item_list_equal(
            validate.ValueIn(
                "a",
                ["b"],
                severity=reports.item.ReportItemSeverity.warning(),
            ).validate({"a": "c"}),
            [
                fixture.warn(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="a",
                    option_value="c",
                    allowed_values=["b"],
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
        )


class ValueInteger(TestCase):
    # The real code only calls ValuePredicateBase and is_integer which are both
    # heavily tested on their own => only basic tests here.
    def fixture_validator(self):
        return validate.ValueInteger("key")

    def test_empty_report_on_valid_option(self):
        assert_report_item_list_equal(
            self.fixture_validator().validate({"key": "2"}), []
        )

    def test_report_invalid_value(self):
        assert_report_item_list_equal(
            self.fixture_validator().validate({"key": "6a"}),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="key",
                    option_value="6a",
                    allowed_values="an integer",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
        )


class ValueIntegerInRange(TestCase):
    # The real code only calls ValuePredicateBase and is_integer which are both
    # heavily tested on their own => only basic tests here.
    def fixture_validator(self):
        return validate.ValueIntegerInRange("key", -5, 5)

    def test_empty_report_on_valid_option(self):
        assert_report_item_list_equal(
            self.fixture_validator().validate({"key": "2"}), []
        )

    def test_report_invalid_value(self):
        assert_report_item_list_equal(
            self.fixture_validator().validate({"key": "6"}),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="key",
                    option_value="6",
                    allowed_values="-5..5",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
        )


class ValueIpAddress(TestCase):
    # The real code only calls ValuePredicateBase and is_ipv4_address and
    # is_ipv6_address which are both heavily tested on their own => only basic
    # tests here.
    def fixture_validator(self):
        return validate.ValueIpAddress("key")

    def test_empty_report_on_ipv4(self):
        assert_report_item_list_equal(
            self.fixture_validator().validate({"key": "192.168.123.42"}), []
        )

    def test_empty_report_on_ipv6(self):
        assert_report_item_list_equal(
            self.fixture_validator().validate({"key": "::192:168:123:42"}), []
        )

    def test_report_invalid_value(self):
        assert_report_item_list_equal(
            self.fixture_validator().validate({"key": "abcd"}),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="key",
                    option_value="abcd",
                    allowed_values="an IP address",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
        )


class ValueNonnegativeInteger(TestCase):
    # The real code only calls ValuePredicateBase => only basic tests here.
    def test_empty_report_on_valid_option(self):
        assert_report_item_list_equal(
            validate.ValueNonnegativeInteger("key").validate({"key": "10"}), []
        )

    def test_report_invalid_value(self):
        assert_report_item_list_equal(
            validate.ValueNonnegativeInteger("key").validate({"key": "-10"}),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="key",
                    option_value="-10",
                    allowed_values="a non-negative integer",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
        )


class ValueNotEmpty(TestCase):
    # The real code only calls ValuePredicateBase => only basic tests here.
    def test_empty_report_on_not_empty_value(self):
        assert_report_item_list_equal(
            validate.ValueNotEmpty("key", "description").validate(
                {"key": "abc"}
            ),
            [],
        )

    def test_empty_report_on_zero_int_value(self):
        assert_report_item_list_equal(
            validate.ValueNotEmpty("key", "description").validate({"key": 0}),
            [],
        )

    def test_report_on_empty_string(self):
        assert_report_item_list_equal(
            validate.ValueNotEmpty("key", "description").validate({"key": ""}),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="key",
                    option_value="",
                    allowed_values="description",
                    cannot_be_empty=True,
                    forbidden_characters=None,
                ),
            ],
        )


class ValuePcmkDatespecPart(TestCase):
    # The real code only calls ValuePredicateBase => only basic tests here.
    def test_empty_report_on_valid_option(self):
        assert_report_item_list_equal(
            validate.ValuePcmkDatespecPart("key", None, None).validate(
                {"key": "10-20"}
            ),
            [],
        )

    def test_report_invalid_value(self):
        assert_report_item_list_equal(
            validate.ValuePcmkDatespecPart("key", 10, 20).validate(
                {"key": "foo-bar"}
            ),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="key",
                    option_value="foo-bar",
                    allowed_values="10..20 or 10..19-11..20",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
        )

    def test_report_invalid_value_no_limits(self):
        assert_report_item_list_equal(
            validate.ValuePcmkDatespecPart("key", None, None).validate(
                {"key": "foo-bar"}
            ),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="key",
                    option_value="foo-bar",
                    allowed_values="an integer or integer-integer",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
        )


class ValuePortNumber(TestCase):
    # The real code only calls ValuePredicateBase => only basic tests here.
    def test_empty_report_on_valid_option(self):
        assert_report_item_list_equal(
            validate.ValuePortNumber("key").validate({"key": "54321"}), []
        )

    def test_report_invalid_value(self):
        assert_report_item_list_equal(
            validate.ValuePortNumber("key").validate({"key": "65536"}),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="key",
                    option_value="65536",
                    allowed_values="a port number (1..65535)",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
        )


class ValuePortRange(TestCase):
    # The real code only calls ValuePredicateBase => only basic tests here.
    def test_empty_report_on_valid_option(self):
        assert_report_item_list_equal(
            validate.ValuePortRange("key").validate({"key": "100-200"}), []
        )

    def test_report_nonsense(self):
        assert_report_item_list_equal(
            validate.ValuePortRange("key").validate({"key": "10-20-30"}),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="key",
                    option_value="10-20-30",
                    allowed_values="port-port",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
        )

    def test_report_bad_start(self):
        assert_report_item_list_equal(
            validate.ValuePortRange("key").validate({"key": "0-100"}),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="key",
                    option_value="0-100",
                    allowed_values="port-port",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
        )

    def test_report_bad_end(self):
        assert_report_item_list_equal(
            validate.ValuePortRange("key").validate({"key": "100-65536"}),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="key",
                    option_value="100-65536",
                    allowed_values="port-port",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
        )


class ValuePositiveInteger(TestCase):
    # The real code only calls ValuePredicateBase => only basic tests here.
    def test_empty_report_on_valid_option(self):
        assert_report_item_list_equal(
            validate.ValuePositiveInteger("key").validate({"key": "10"}), []
        )

    def test_report_invalid_value(self):
        assert_report_item_list_equal(
            validate.ValuePositiveInteger("key").validate({"key": "0"}),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="key",
                    option_value="0",
                    allowed_values="a positive integer",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
        )


class ValueScore(TestCase):
    def test_valid_score(self):
        for score in [
            "1",
            "-1",
            "+1",
            "123",
            "-123",
            "+123",
            "INFINITY",
            "-INFINITY",
            "+INFINITY",
        ]:
            with self.subTest(score=score):
                assert_report_item_list_equal(
                    validate.ValueScore("a").validate({"a": score}),
                    [],
                )

    def test_not_valid_score(self):
        for score in ["something", "++1", "--1", "++INFINITY"]:
            with self.subTest(score=score):
                assert_report_item_list_equal(
                    validate.ValueScore("a").validate({"a": score}),
                    [
                        fixture.error(
                            reports.codes.INVALID_SCORE,
                            score=score,
                        ),
                    ],
                )


class ValueTimeInterval(TestCase):
    def test_no_reports_for_valid_time_interval(self):
        for interval in ["0", "1s", "2sec", "3m", "4min", "5h", "6hr"]:
            with self.subTest(value=interval):
                assert_report_item_list_equal(
                    validate.ValueTimeInterval("a").validate({"a": interval}),
                    [],
                )

    def test_reports_about_invalid_interval(self):
        assert_report_item_list_equal(
            validate.ValueTimeInterval("a").validate({"a": "invalid_value"}),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="a",
                    option_value="invalid_value",
                    allowed_values="time interval (e.g. 1, 2s, 3m, 4h, ...)",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
        )


class ValueVersion(TestCase):
    def test_no_reports_for_valid_time_interval(self):
        for version in ["123", "123.456", "123.456.789", "1.2.3.4"]:
            with self.subTest(value=version):
                assert_report_item_list_equal(
                    validate.ValueVersion("a").validate({"a": version}),
                    [],
                )

    def test_reports_about_invalid_interval(self):
        assert_report_item_list_equal(
            validate.ValueVersion("a").validate({"a": "1.2.3a"}),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="a",
                    option_value="1.2.3a",
                    allowed_values="a version number (e.g. 1, 1.2, 1.23.45, ...)",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
        )


### predicates


class IsFloat(TestCase):
    def test_success(self):
        self.assertTrue(validate.is_float(123))
        self.assertTrue(validate.is_float("123"))
        self.assertTrue(validate.is_float(-123))
        self.assertTrue(validate.is_float("-123"))
        self.assertTrue(validate.is_float(+123))
        self.assertTrue(validate.is_float("+123"))

        self.assertTrue(validate.is_float("123.098"))
        self.assertTrue(validate.is_float(".123"))
        self.assertTrue(validate.is_float("123."))

        self.assertTrue(validate.is_float("123e123"))
        self.assertTrue(validate.is_float("123E123"))
        self.assertTrue(validate.is_float("123e+123"))
        self.assertTrue(validate.is_float("123E+123"))
        self.assertTrue(validate.is_float("+123e123"))
        self.assertTrue(validate.is_float("+123E123"))
        self.assertTrue(validate.is_float("+123e+123"))
        self.assertTrue(validate.is_float("+123E+123"))
        self.assertTrue(validate.is_float("123e-123"))
        self.assertTrue(validate.is_float("123E-123"))
        self.assertTrue(validate.is_float("-123e123"))
        self.assertTrue(validate.is_float("-123E123"))
        self.assertTrue(validate.is_float("-123e-123"))
        self.assertTrue(validate.is_float("-123E-123"))
        self.assertTrue(validate.is_float("-12.3e-123"))
        self.assertTrue(validate.is_float("-12.3E-123"))

        self.assertFalse(validate.is_float(" 1"))
        self.assertFalse(validate.is_float("12_34"))
        self.assertFalse(validate.is_float("\n-1"))
        self.assertFalse(validate.is_float("\r+1"))
        self.assertFalse(validate.is_float("1\n"))
        self.assertFalse(validate.is_float("-1 "))
        self.assertFalse(validate.is_float("+1\r"))

        self.assertFalse(validate.is_float(""))
        self.assertFalse(validate.is_float("1a"))
        self.assertFalse(validate.is_float("a1"))
        self.assertFalse(validate.is_float("aaa"))


class IsInteger(TestCase):
    def test_no_range(self):
        self.assertTrue(validate.is_integer(1))
        self.assertTrue(validate.is_integer("1"))
        self.assertTrue(validate.is_integer(-1))
        self.assertTrue(validate.is_integer("-1"))
        self.assertTrue(validate.is_integer(+1))
        self.assertTrue(validate.is_integer("+1"))

        self.assertFalse(validate.is_integer(" 1"))
        self.assertFalse(validate.is_integer("\n-1"))
        self.assertFalse(validate.is_integer("\r+1"))
        self.assertFalse(validate.is_integer("1\n"))
        self.assertFalse(validate.is_integer("-1 "))
        self.assertFalse(validate.is_integer("+1\r"))

        self.assertFalse(validate.is_integer(""))
        self.assertFalse(validate.is_integer("1a"))
        self.assertFalse(validate.is_integer("a1"))
        self.assertFalse(validate.is_integer("aaa"))
        self.assertFalse(validate.is_integer(1.0))
        self.assertFalse(validate.is_integer("1.0"))

    def test_at_least(self):
        self.assertTrue(validate.is_integer(5, 5))
        self.assertTrue(validate.is_integer(5, 4))
        self.assertTrue(validate.is_integer("5", 5))
        self.assertTrue(validate.is_integer("5", 4))

        self.assertFalse(validate.is_integer(5, 6))
        self.assertFalse(validate.is_integer("5", 6))

    def test_at_most(self):
        self.assertTrue(validate.is_integer(5, None, 5))
        self.assertTrue(validate.is_integer(5, None, 6))
        self.assertTrue(validate.is_integer("5", None, 5))
        self.assertTrue(validate.is_integer("5", None, 6))

        self.assertFalse(validate.is_integer(5, None, 4))
        self.assertFalse(validate.is_integer("5", None, 4))

    def test_range(self):
        self.assertTrue(validate.is_integer(5, 5, 5))
        self.assertTrue(validate.is_integer(5, 4, 6))
        self.assertTrue(validate.is_integer("5", 5, 5))
        self.assertTrue(validate.is_integer("5", 4, 6))

        self.assertFalse(validate.is_integer(3, 4, 6))
        self.assertFalse(validate.is_integer(7, 4, 6))
        self.assertFalse(validate.is_integer("3", 4, 6))
        self.assertFalse(validate.is_integer("7", 4, 6))


class IsIpv4Address(TestCase):
    def test_valid(self):
        self.assertTrue(validate.is_ipv4_address("192.168.1.1"))
        self.assertTrue(validate.is_ipv4_address("1.2.3.4"))
        self.assertTrue(validate.is_ipv4_address("255.255.255.255"))
        self.assertTrue(validate.is_ipv4_address("0.0.0.0"))

    def test_bad(self):
        self.assertFalse(validate.is_ipv4_address("abcd"))
        self.assertFalse(validate.is_ipv4_address("192 168 1 1"))
        self.assertFalse(validate.is_ipv4_address("3232235521"))
        self.assertFalse(validate.is_ipv4_address("::1"))
        self.assertFalse(validate.is_ipv4_address(1234))


class IsIpv6Address(TestCase):
    def test_valid(self):
        self.assertTrue(validate.is_ipv6_address("fe80::5054:ff:fec6:8eaf"))
        self.assertTrue(validate.is_ipv6_address("::abc:7:def"))

    def test_bad(self):
        self.assertFalse(validate.is_ipv6_address("abcd"))
        self.assertFalse(validate.is_ipv6_address("192.168.1.1"))
        self.assertFalse(validate.is_ipv6_address(1234))


class IsPcmkDatespecPart(TestCase):
    def test_valid(self):
        self.assertTrue(validate.is_pcmk_datespec_part("12"))
        self.assertTrue(validate.is_pcmk_datespec_part("12-34"))

    def test_valid_with_limits(self):
        self.assertTrue(validate.is_pcmk_datespec_part("12", 10, 20))
        self.assertTrue(validate.is_pcmk_datespec_part("12-13", 10, 20))

    def test_bad_not_int(self):
        self.assertFalse(validate.is_pcmk_datespec_part("foo"))
        self.assertFalse(validate.is_pcmk_datespec_part("1-foo"))
        self.assertFalse(validate.is_pcmk_datespec_part("foo-1"))

    def test_bad_since_after_until(self):
        self.assertFalse(validate.is_pcmk_datespec_part("10-10"))
        self.assertFalse(validate.is_pcmk_datespec_part("10-9"))

    def test_bad_limits(self):
        self.assertFalse(validate.is_pcmk_datespec_part("5", 10, 20))
        self.assertFalse(validate.is_pcmk_datespec_part("5-15", 10, 20))
        self.assertFalse(validate.is_pcmk_datespec_part("15-25", 10, 20))


class IsPortNumber(TestCase):
    def test_valid_port(self):
        self.assertTrue(validate.is_port_number(1))
        self.assertTrue(validate.is_port_number("1"))
        self.assertTrue(validate.is_port_number(65535))
        self.assertTrue(validate.is_port_number("65535"))
        self.assertTrue(validate.is_port_number(8192))

    def test_bad_port(self):
        self.assertFalse(validate.is_port_number(0))
        self.assertFalse(validate.is_port_number("0"))
        self.assertFalse(validate.is_port_number(65536))
        self.assertFalse(validate.is_port_number("65536"))
        self.assertFalse(validate.is_port_number(" 8192 "))
        self.assertFalse(validate.is_port_number(-128))
        self.assertFalse(validate.is_port_number("-128"))
        self.assertFalse(validate.is_port_number("abcd"))


class MatchesRegexp(TestCase):
    def test_matches_string(self):
        self.assertTrue(validate.matches_regexp("abcdcba", "^[a-d]+$"))

    def test_matches_regexp(self):
        self.assertTrue(
            validate.matches_regexp(
                "abCDCBa", re.compile("^[a-d]+$", re.IGNORECASE)
            )
        )

    def test_not_matches_string(self):
        self.assertFalse(validate.matches_regexp("abcDcba", "^[a-d]+$"))

    def test_not_matches_regexp(self):
        self.assertFalse(
            validate.matches_regexp(
                "abCeCBa", re.compile("^[a-d]+$", re.IGNORECASE)
            )
        )


class IsEmptyString(TestCase):
    def test_empty_string(self):
        self.assertTrue(validate.is_empty_string(""))

    def test_not_empty_string(self):
        self.assertFalse(validate.is_empty_string("a"))
        self.assertFalse(validate.is_empty_string("0"))
        self.assertFalse(validate.is_empty_string(0))



import random
import string
from time import time as now


class Session:
    def __init__(
        self,
        sid,
        username=None,
        groups=None,
        is_authenticated=False,
        ajax_id=None,
    ):
        # Session id propageted via cookies.
        self.__sid = sid
        # Username given by login attempt. It does not matter if the
        # authentication succeeded.
        self.__username = username
        # The flag that indicated if the user was authenticated.
        # The user is authenticated when they are recognized as a system user
        # belonging to the high availability admin group (typicaly hacluster).
        self.__is_authenticated = is_authenticated
        # Id that will be returned by login-status or login (for ajax).
        self.__ajax_id = None
        if self.__is_authenticated:
            self.__ajax_id = (
                ajax_id if ajax_id else f"{int(now())}-{random.randint(1, 100)}"
            )
        # Groups of the user. Similary to username, it does not mean that the
        # user is authenticated when the groups are loaded.
        self.__groups = groups or []
        # The moment of the last access. The only muttable attribute.
        self.refresh()

    @property
    def is_authenticated(self):
        self.refresh()
        return self.__is_authenticated

    @property
    def username(self):
        self.refresh()
        return self.__username

    @property
    def sid(self):
        self.refresh()
        return self.__sid

    @property
    def ajax_id(self):
        self.refresh()
        return self.__ajax_id

    @property
    def groups(self):
        self.refresh()
        return self.__groups

    def refresh(self):
        """
        Set the time of last access to now.
        """
        self.__last_access = now()
        return self

    def was_unused_last(self, seconds):
        return now() > self.__last_access + seconds


class Storage:
    def __init__(self, lifetime_seconds):
        self.__sessions = {}
        self.__lifetime_seconds = lifetime_seconds

    def provide(self, sid=None) -> Session:
        if self.__is_valid_sid(sid):
            return self.__sessions[sid].refresh()
        return self.__register(self.__generate_sid())

    def drop_expired(self):
        obsolete_sid_list = [
            sid
            for sid, session in self.__sessions.items()
            if session.was_unused_last(self.__lifetime_seconds)
        ]
        for sid in obsolete_sid_list:
            del self.__sessions[sid]

    def destroy(self, sid):
        if sid in self.__sessions:
            del self.__sessions[sid]
        return self

    def login(self, sid, username, groups, ajax_id=None) -> Session:
        return self.__register(
            self.__valid_sid(sid),
            username=username,
            groups=groups,
            is_authenticated=True,
            ajax_id=ajax_id,
        )

    def rejected_user(self, sid, username) -> Session:
        return self.__register(self.__valid_sid(sid), username=username)

    def __is_valid_sid(self, sid):
        return not (
            sid is None
            or
            # Do not let a user (an attacker?) to force us to use their sid.
            sid not in self.__sessions
            or self.__sessions[sid].was_unused_last(self.__lifetime_seconds)
        )

    def __valid_sid(self, sid):
        return sid if self.__is_valid_sid(sid) else self.__generate_sid()

    def __register(self, *args, **kwargs) -> Session:
        session = Session(*args, **kwargs)
        self.__sessions[session.sid] = session
        return session

    def __generate_sid(self):
        for _ in range(10):
            sid = "".join(
                random.choices(string.ascii_lowercase + string.digits, k=64)
            )
            if sid not in self.__sessions:
                return sid
        # TODO what to do?
        raise Exception("Cannot generate unique sid")



import re
from textwrap import dedent
from unittest import TestCase

from pcs_test.tools import fixture
from pcs_test.tools.assertions import (
    ac,
    assert_raise_library_error,
)
from pcs_test.tools.misc import get_test_resource as rc, outdent

from pcs.common.reports import ReportItemSeverity as severity
from pcs.common.reports import codes as report_codes
import pcs.lib.corosync.config_facade as lib
from pcs.lib.corosync.config_parser import Parser


def _read_file(name):
    with open(rc(name)) as a_file:
        return a_file.read()


def _get_facade(config_text):
    return lib.ConfigFacade(Parser.parse(config_text.encode("utf-8")))


class HasQuorumDeviceTest(TestCase):
    def test_empty_config(self):
        config = ""
        facade = _get_facade(config)
        self.assertFalse(facade.has_quorum_device())
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)

    def test_no_device(self):
        config = _read_file("corosync.conf")
        facade = _get_facade(config)
        self.assertFalse(facade.has_quorum_device())
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)

    def test_empty_device(self):
        config = dedent(
            """\
            quorum {
                device {
                }
            }
        """
        )
        facade = _get_facade(config)
        self.assertFalse(facade.has_quorum_device())
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)

    def test_device_set(self):
        config = dedent(
            """\
            quorum {
                device {
                    model: net
                }
            }
        """
        )
        facade = _get_facade(config)
        self.assertTrue(facade.has_quorum_device())
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)

    def test_no_model(self):
        config = dedent(
            """\
            quorum {
                device {
                    option: value
                    net {
                        host: 127.0.0.1
                    }
                }
            }
        """
        )
        facade = _get_facade(config)
        self.assertFalse(facade.has_quorum_device())
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)


class GetQuorumDeviceModel(TestCase):
    def assert_model(self, config, expected_model):
        facade = _get_facade(config)
        self.assertEqual(expected_model, facade.get_quorum_device_model())
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)

    def test_empty_config(self):
        config = ""
        self.assert_model(config, None)

    def test_no_device(self):
        config = _read_file("corosync.conf")
        self.assert_model(config, None)

    def test_empty_device(self):
        config = dedent(
            """\
            quorum {
                device {
                }
            }
        """
        )
        self.assert_model(config, None)

    def test_no_model(self):
        config = dedent(
            """\
            quorum {
                device {
                    option: value
                    net {
                        host: 127.0.0.1
                    }
                }
            }
        """
        )
        self.assert_model(config, None)

    def test_configured_properly(self):
        config = dedent(
            """\
            quorum {
                device {
                    option: value
                    model: net
                    net {
                        host: 127.0.0.1
                    }
                }
            }
        """
        )
        self.assert_model(config, "net")

    def test_more_devices_one_quorum(self):
        config = dedent(
            """\
            quorum {
                device {
                    option0: valueX
                    option1: value1
                    model: disk
                    net {
                        host: 127.0.0.1
                    }
                    heuristics {
                        mode: sync
                        exec_ls: test -f /tmp/test
                    }
                }
                device {
                    option0: valueY
                    option2: value2
                    model: net
                    disk {
                        path: /dev/quorum_disk
                    }
                    heuristics {
                        mode: on
                    }
                    heuristics {
                        timeout: 5
                    }
                }
            }
        """
        )
        self.assert_model(config, "net")

    def test_more_devices_more_quorum(self):
        config = dedent(
            """\
            quorum {
                device {
                    option0: valueX
                    option1: value1
                    model: disk
                    net {
                        host: 127.0.0.1
                    }
                    heuristics {
                        mode: sync
                        exec_ls: test -f /tmp/test
                    }
                }
            }
            quorum {
                device {
                    option0: valueY
                    option2: value2
                    model: net
                    disk {
                        path: /dev/quorum_disk
                    }
                    heuristics {
                        mode: on
                    }
                    heuristics {
                        timeout: 5
                    }
                }
            }
        """
        )
        self.assert_model(config, "net")


class GetQuorumDeviceSettingsTest(TestCase):
    def test_empty_config(self):
        config = ""
        facade = _get_facade(config)
        self.assertEqual(
            (None, {}, {}, {}), facade.get_quorum_device_settings()
        )
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)

    def test_no_device(self):
        config = _read_file("corosync.conf")
        facade = _get_facade(config)
        self.assertEqual(
            (None, {}, {}, {}), facade.get_quorum_device_settings()
        )
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)

    def test_empty_device(self):
        config = dedent(
            """\
            quorum {
                device {
                }
            }
        """
        )
        facade = _get_facade(config)
        self.assertEqual(
            (None, {}, {}, {}), facade.get_quorum_device_settings()
        )
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)

    def test_no_model(self):
        config = dedent(
            """\
            quorum {
                device {
                    option: value
                    net {
                        host: 127.0.0.1
                    }
                }
            }
        """
        )
        facade = _get_facade(config)
        self.assertEqual(
            (None, {}, {"option": "value"}, {}),
            facade.get_quorum_device_settings(),
        )
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)

    def test_configured_properly(self):
        config = dedent(
            """\
            quorum {
                device {
                    option: value
                    model: net
                    net {
                        host: 127.0.0.1
                    }
                }
            }
        """
        )
        facade = _get_facade(config)
        self.assertEqual(
            ("net", {"host": "127.0.0.1"}, {"option": "value"}, {}),
            facade.get_quorum_device_settings(),
        )
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)

    def test_configured_properly_heuristics(self):
        config = dedent(
            """\
            quorum {
                device {
                    option: value
                    model: net
                    net {
                        host: 127.0.0.1
                    }
                    heuristics {
                        mode: on
                        exec_ls: test -f /tmp/test
                    }
                }
            }
        """
        )
        facade = _get_facade(config)
        self.assertEqual(
            (
                "net",
                {"host": "127.0.0.1"},
                {"option": "value"},
                {"exec_ls": "test -f /tmp/test", "mode": "on"},
            ),
            facade.get_quorum_device_settings(),
        )
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)

    def test_more_devices_one_quorum(self):
        config = dedent(
            """\
            quorum {
                device {
                    option0: valueX
                    option1: value1
                    model: disk
                    net {
                        host: 127.0.0.1
                    }
                    heuristics {
                        mode: sync
                        exec_ls: test -f /tmp/test
                    }
                }
                device {
                    option0: valueY
                    option2: value2
                    model: net
                    disk {
                        path: /dev/quorum_disk
                    }
                    heuristics {
                        mode: on
                    }
                    heuristics {
                        timeout: 5
                    }
                }
            }
        """
        )
        facade = _get_facade(config)
        self.assertEqual(
            (
                "net",
                {"host": "127.0.0.1"},
                {"option0": "valueY", "option1": "value1", "option2": "value2"},
                {"exec_ls": "test -f /tmp/test", "mode": "on", "timeout": "5"},
            ),
            facade.get_quorum_device_settings(),
        )
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)

    def test_more_devices_more_quorum(self):
        config = dedent(
            """\
            quorum {
                device {
                    option0: valueX
                    option1: value1
                    model: disk
                    net {
                        host: 127.0.0.1
                    }
                    heuristics {
                        mode: sync
                        exec_ls: test -f /tmp/test
                    }
                }
            }
            quorum {
                device {
                    option0: valueY
                    option2: value2
                    model: net
                    disk {
                        path: /dev/quorum_disk
                    }
                    heuristics {
                        mode: on
                    }
                    heuristics {
                        timeout: 5
                    }
                }
            }
        """
        )
        facade = _get_facade(config)
        self.assertEqual(
            (
                "net",
                {"host": "127.0.0.1"},
                {"option0": "valueY", "option1": "value1", "option2": "value2"},
                {"exec_ls": "test -f /tmp/test", "mode": "on", "timeout": "5"},
            ),
            facade.get_quorum_device_settings(),
        )
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)


class IsQuorumDeviceHeuristicsEnabledWithNoExec(TestCase):
    def assert_result(self, config, expected_result):
        facade = _get_facade(config)
        self.assertEqual(
            expected_result,
            facade.is_quorum_device_heuristics_enabled_with_no_exec(),
        )
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)

    def test_empty_config(self):
        config = ""
        self.assert_result(config, False)

    def test_no_device(self):
        config = _read_file("corosync.conf")
        self.assert_result(config, False)

    def test_empty_heuristics(self):
        config = dedent(
            """\
            quorum {
                device {
                    heuristics {
                    }
                }
            }
        """
        )
        self.assert_result(config, False)

    def test_heuristics_off(self):
        config = dedent(
            """\
            quorum {
                device {
                    heuristics {
                        mode: off
                    }
                }
            }
        """
        )
        self.assert_result(config, False)

    def test_heuristics_on_no_exec(self):
        config = dedent(
            """\
            quorum {
                device {
                    heuristics {
                        mode: on
                    }
                }
            }
        """
        )
        self.assert_result(config, True)

    def test_heuristics_sync_no_exec(self):
        config = dedent(
            """\
            quorum {
                device {
                    heuristics {
                        mode: on
                        exec_ls:
                    }
                }
            }
        """
        )
        self.assert_result(config, True)

    def test_heuristics_on_with_exec(self):
        config = dedent(
            """\
            quorum {
                device {
                    heuristics {
                        mode: on
                        exec_ls: test -f /tmp/test
                    }
                }
            }
        """
        )
        self.assert_result(config, False)

    def test_heuristics_sync_with_exec(self):
        config = dedent(
            """\
            quorum {
                device {
                    heuristics {
                        mode: sync
                        exec_ls: test -f /tmp/test
                    }
                }
            }
        """
        )
        self.assert_result(config, False)

    def test_heuristics_unknown_no_exec(self):
        config = dedent(
            """\
            quorum {
                device {
                    heuristics {
                        mode: unknown
                    }
                }
            }
        """
        )
        self.assert_result(config, False)

    def test_heuristics_unknown_with_exec(self):
        config = dedent(
            """\
            quorum {
                device {
                    heuristics {
                        mode: unknown
                    }
                }
            }
        """
        )
        self.assert_result(config, False)


class AddQuorumDeviceTest(TestCase):
    def test_success_net_minimal_ffsplit(self):
        config = _read_file("corosync-3nodes.conf")
        facade = _get_facade(config)
        facade.add_quorum_device(
            "net", {"host": "127.0.0.1", "algorithm": "ffsplit"}, {}, {}
        )
        ac(
            config.replace(
                "    provider: corosync_votequorum\n",
                outdent(
                    """\
                    provider: corosync_votequorum

                    device {
                        model: net
                        votes: 1

                        net {
                            algorithm: ffsplit
                            host: 127.0.0.1
                        }
                    }
                """
                ),
            ),
            facade.config.export(),
        )
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)

    def test_success_net_minimal_lms(self):
        config = _read_file("corosync-3nodes.conf")
        facade = _get_facade(config)
        facade.add_quorum_device(
            "net", {"host": "127.0.0.1", "algorithm": "lms"}, {}, {}
        )
        ac(
            config.replace(
                "    provider: corosync_votequorum\n",
                outdent(
                    """\
                    provider: corosync_votequorum

                    device {
                        model: net

                        net {
                            algorithm: lms
                            host: 127.0.0.1
                        }
                    }
                """
                ),
            ),
            facade.config.export(),
        )
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)

    def test_success_remove_nodes_votes(self):
        config = _read_file("corosync-3nodes.conf")
        config_votes = config.replace("node {", "node {\nquorum_votes: 2")
        facade = _get_facade(config_votes)
        facade.add_quorum_device(
            "net", {"host": "127.0.0.1", "algorithm": "lms"}, {}, {}
        )
        ac(
            config.replace(
                "    provider: corosync_votequorum\n",
                outdent(
                    """\
                    provider: corosync_votequorum

                    device {
                        model: net

                        net {
                            algorithm: lms
                            host: 127.0.0.1
                        }
                    }
                """
                ),
            ),
            facade.config.export(),
        )
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)

    def test_success_net_full(self):
        config = _read_file("corosync-3nodes.conf")
        facade = _get_facade(config)
        facade.add_quorum_device(
            "net",
            {
                "host": "127.0.0.1",
                "port": "4433",
                "algorithm": "ffsplit",
                "connect_timeout": "12345",
                "force_ip_version": "4",
                "tie_breaker": "lowest",
            },
            {"timeout": "23456", "sync_timeout": "34567"},
            {
                "mode": "on",
                "timeout": "5",
                "sync_timeout": "15",
                "interval": "30",
                "exec_ping": 'ping -q -c 1 "127.0.0.1"',
                "exec_ls": "test -f /tmp/test",
            },
        )
        ac(
            config.replace(
                "    provider: corosync_votequorum\n",
                outdent(
                    """\
                    provider: corosync_votequorum

                    device {
                        sync_timeout: 34567
                        timeout: 23456
                        model: net
                        votes: 1

                        net {
                            algorithm: ffsplit
                            connect_timeout: 12345
                            force_ip_version: 4
                            host: 127.0.0.1
                            port: 4433
                            tie_breaker: lowest
                        }

                        heuristics {
                            exec_ls: test -f /tmp/test
                            exec_ping: ping -q -c 1 "127.0.0.1"
                            interval: 30
                            mode: on
                            sync_timeout: 15
                            timeout: 5
                        }
                    }
                """
                ),
            ),
            facade.config.export(),
        )
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)

    def test_remove_conflicting_options(self):
        config = _read_file("corosync.conf")
        config = config.replace(
            "    two_node: 1\n",
            "\n".join(
                [
                    "    two_node: 1",
                    "    auto_tie_breaker: 1",
                    "    last_man_standing: 1",
                    "    last_man_standing_window: 987",
                    "    allow_downscale: 1",
                    "",
                ]
            ),
        )
        facade = _get_facade(config)
        facade.add_quorum_device(
            "net", {"host": "127.0.0.1", "algorithm": "ffsplit"}, {}, {}
        )
        ac(
            re.sub(
                re.compile(r"quorum {[^}]*}\n", re.MULTILINE | re.DOTALL),
                dedent(
                    """\
                    quorum {
                        provider: corosync_votequorum

                        device {
                            model: net
                            votes: 1

                            net {
                                algorithm: ffsplit
                                host: 127.0.0.1
                            }
                        }
                    }
                """
                ),
                config,
            ),
            facade.config.export(),
        )
        self.assertTrue(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)

    def test_remove_old_configuration(self):
        config = dedent(
            """\
            quorum {
                provider: corosync_votequorum
                device {
                    option: value_old1
                    heuristics {
                        h_option: hvalue_old1
                    }
                }
            }
            quorum {
                provider: corosync_votequorum
                device {
                    option: value_old2
                    heuristics {
                        h_option: hvalue_old2
                    }
                }
            }
        """
        )
        facade = _get_facade(config)
        facade.add_quorum_device(
            "net", {"host": "127.0.0.1", "algorithm": "ffsplit"}, {}, {}
        )
        ac(
            dedent(
                """\
                quorum {
                    provider: corosync_votequorum
                }

                quorum {
                    provider: corosync_votequorum

                    device {
                        model: net
                        votes: 1

                        net {
                            algorithm: ffsplit
                            host: 127.0.0.1
                        }
                    }
                }
            """
            ),
            facade.config.export(),
        )
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)


class UpdateQuorumDeviceTest(TestCase):
    @staticmethod
    def fixture_add_device(config, votes=None):
        with_device = re.sub(
            re.compile(r"quorum {[^}]*}", re.MULTILINE | re.DOTALL),
            dedent(
                """\
                quorum {
                    provider: corosync_votequorum

                    device {
                        timeout: 12345
                        model: net

                        net {
                            host: 127.0.0.1
                            port: 4433
                        }
                    }
                }"""
            ),
            config,
        )
        if votes:
            with_device = with_device.replace(
                "model: net", "model: net\n        votes: {0}".format(votes)
            )
        return with_device

    @staticmethod
    def fixture_add_device_with_heuristics(config, votes=None):
        with_device = re.sub(
            re.compile(r"quorum {[^}]*}", re.MULTILINE | re.DOTALL),
            dedent(
                """\
                quorum {
                    provider: corosync_votequorum

                    device {
                        timeout: 12345
                        model: net

                        net {
                            host: 127.0.0.1
                            port: 4433
                        }

                        heuristics {
                            exec_ls: test -f /tmp/test
                            interval: 30
                            mode: on
                        }
                    }
                }"""
            ),
            config,
        )
        if votes:
            with_device = with_device.replace(
                "model: net", "model: net\n        votes: {0}".format(votes)
            )
        return with_device

    def test_not_existing(self):
        config = _read_file("corosync.conf")
        facade = _get_facade(config)
        assert_raise_library_error(
            lambda: facade.update_quorum_device({"host": "127.0.0.1"}, {}, {}),
            (severity.ERROR, report_codes.QDEVICE_NOT_DEFINED, {}),
        )
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)
        ac(config, facade.config.export())

    def test_not_existing_add_heuristics(self):
        config = _read_file("corosync.conf")
        facade = _get_facade(config)
        assert_raise_library_error(
            lambda: facade.update_quorum_device({}, {}, {"mode": "on"}),
            fixture.error(report_codes.QDEVICE_NOT_DEFINED),
        )
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)
        ac(config, facade.config.export())

    def test_success_model_options_net(self):
        config = self.fixture_add_device(
            _read_file("corosync-3nodes.conf"), votes="1"
        )
        facade = _get_facade(config)
        facade.update_quorum_device(
            {"host": "127.0.0.2", "port": "", "algorithm": "ffsplit"}, {}, {}
        )
        self.assertFalse(facade.need_stopped_cluster)
        self.assertTrue(facade.need_qdevice_reload)
        ac(
            config.replace(
                "host: 127.0.0.1\n            port: 4433",
                "host: 127.0.0.2\n            algorithm: ffsplit",
            ),
            facade.config.export(),
        )

    def test_success_generic_options(self):
        config = self.fixture_add_device(_read_file("corosync-3nodes.conf"))
        facade = _get_facade(config)
        facade.update_quorum_device(
            {}, {"timeout": "", "sync_timeout": "23456"}, {}
        )
        self.assertFalse(facade.need_stopped_cluster)
        self.assertTrue(facade.need_qdevice_reload)
        ac(
            config.replace(
                "timeout: 12345\n        model: net",
                "model: net\n        sync_timeout: 23456",
            ),
            facade.config.export(),
        )

    def test_success_all_options(self):
        config = self.fixture_add_device_with_heuristics(
            _read_file("corosync-3nodes.conf")
        )
        facade = _get_facade(config)
        facade.update_quorum_device(
            {"port": "4444"}, {"timeout": "23456"}, {"interval": "35"}
        )
        self.assertFalse(facade.need_stopped_cluster)
        self.assertTrue(facade.need_qdevice_reload)
        ac(
            config.replace("port: 4433", "port: 4444")
            .replace("timeout: 12345", "timeout: 23456")
            .replace("interval: 30", "interval: 35"),
            facade.config.export(),
        )

    def test_success_add_heuristics(self):
        config = self.fixture_add_device(_read_file("corosync-3nodes.conf"))
        facade = _get_facade(config)
        facade.update_quorum_device(
            {},
            {},
            {"mode": "on", "exec_ls": "test -f /tmp/test", "interval": "30"},
        )
        self.assertFalse(facade.need_stopped_cluster)
        self.assertTrue(facade.need_qdevice_reload)
        ac(
            self.fixture_add_device_with_heuristics(
                _read_file("corosync-3nodes.conf")
            ),
            facade.config.export(),
        )

    def test_success_remove_heuristics(self):
        config = self.fixture_add_device_with_heuristics(
            _read_file("corosync-3nodes.conf")
        )
        facade = _get_facade(config)
        facade.update_quorum_device(
            {}, {}, {"mode": "", "exec_ls": "", "interval": ""}
        )
        self.assertFalse(facade.need_stopped_cluster)
        self.assertTrue(facade.need_qdevice_reload)
        ac(
            self.fixture_add_device(_read_file("corosync-3nodes.conf")),
            facade.config.export(),
        )

    def test_success_change_heuristics(self):
        config = self.fixture_add_device_with_heuristics(
            _read_file("corosync-3nodes.conf")
        )
        facade = _get_facade(config)
        facade.update_quorum_device(
            {}, {}, {"mode": "sync", "interval": "", "timeout": "20"}
        )
        self.assertFalse(facade.need_stopped_cluster)
        self.assertTrue(facade.need_qdevice_reload)
        ac(
            config.replace(
                "interval: 30\n            mode: on",
                "mode: sync\n            timeout: 20",
            ),
            facade.config.export(),
        )


class RemoveQuorumDeviceTest(TestCase):
    def test_empty_config(self):
        config = ""
        facade = _get_facade(config)
        assert_raise_library_error(
            facade.remove_quorum_device,
            (severity.ERROR, report_codes.QDEVICE_NOT_DEFINED, {}),
        )
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)

    def test_no_device(self):
        config = _read_file("corosync-3nodes.conf")
        facade = _get_facade(config)
        assert_raise_library_error(
            facade.remove_quorum_device,
            (severity.ERROR, report_codes.QDEVICE_NOT_DEFINED, {}),
        )
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)

    def test_remove_all_devices(self):
        config_no_devices = _read_file("corosync-3nodes.conf")
        config = re.sub(
            re.compile(r"quorum {[^}]*}", re.MULTILINE | re.DOTALL),
            dedent(
                """\
                quorum {
                    provider: corosync_votequorum

                    device {
                        option: value
                        model: net

                        net {
                            host: 127.0.0.1
                            port: 4433
                        }
                    }

                    device {
                        option: value
                    }
                }

                quorum {
                    device {
                        option: value
                        model: disk

                        net {
                            host: 127.0.0.1
                            port: 4433
                        }
                    }
                }
            """
            ),
            config_no_devices,
        )
        facade = _get_facade(config)
        facade.remove_quorum_device()
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)
        ac(config_no_devices, facade.config.export())

    def test_restore_two_node(self):
        config_no_devices = _read_file("corosync.conf")
        config = re.sub(
            re.compile(r"quorum {[^}]*}", re.MULTILINE | re.DOTALL),
            dedent(
                """\
                quorum {
                    provider: corosync_votequorum

                    device {
                        option: value
                        model: net

                        net {
                            host: 127.0.0.1
                            port: 4433
                        }
                    }
                }
            """
            ),
            config_no_devices,
        )
        facade = _get_facade(config)
        facade.remove_quorum_device()
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)
        ac(config_no_devices, facade.config.export())


class RemoveQuorumDeviceHeuristics(TestCase):
    def test_error_on_empty_config(self):
        config = ""
        facade = _get_facade(config)
        assert_raise_library_error(
            facade.remove_quorum_device_heuristics,
            fixture.error(report_codes.QDEVICE_NOT_DEFINED),
        )
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)

    def test_error_on_no_device(self):
        config = _read_file("corosync-3nodes.conf")
        facade = _get_facade(config)
        assert_raise_library_error(
            facade.remove_quorum_device_heuristics,
            fixture.error(report_codes.QDEVICE_NOT_DEFINED),
        )
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)

    def test_noop_on_no_heuristics(self):
        config = _read_file("corosync-3nodes-qdevice.conf")
        facade = _get_facade(config)
        facade.remove_quorum_device_heuristics()
        self.assertFalse(facade.need_stopped_cluster)
        self.assertTrue(facade.need_qdevice_reload)
        ac(config, facade.config.export())

    def test_remove_all_heuristics(self):
        config_no_devices = _read_file("corosync-3nodes.conf")
        config_no_heuristics = re.sub(
            re.compile(r"quorum {[^}]*}", re.MULTILINE | re.DOTALL),
            dedent(
                """\
                quorum {
                    provider: corosync_votequorum

                    device {
                        model: net

                        net {
                            host: 127.0.0.1
                        }
                    }

                    device {
                        option: value
                    }
                }

                quorum {
                    device {
                        model: net

                        net {
                            host: 127.0.0.2
                        }
                    }
                }"""
            ),
            config_no_devices,
        )
        config_heuristics = re.sub(
            re.compile(r"quorum {[^}]*}", re.MULTILINE | re.DOTALL),
            dedent(
                """\
                quorum {
                    provider: corosync_votequorum

                    device {
                        model: net

                        net {
                            host: 127.0.0.1
                        }

                        heuristics {
                            mode: on
                        }
                    }

                    device {
                        option: value

                        heuristics {
                            interval: 3000
                        }
                    }
                }

                quorum {
                    device {
                        model: net

                        net {
                            host: 127.0.0.2
                        }

                        heuristics {
                            exec_ls: test -f /tmp/test
                        }
                    }
                }"""
            ),
            config_no_devices,
        )

        facade = _get_facade(config_heuristics)
        facade.remove_quorum_device_heuristics()
        self.assertFalse(facade.need_stopped_cluster)
        self.assertTrue(facade.need_qdevice_reload)
        ac(config_no_heuristics, facade.config.export())



import os.path
from unittest import TestCase

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools

from pcs import settings
from pcs.common.reports import codes as report_codes
from pcs.lib.commands import cluster


class SuccessMinimal(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.nodes = [f"node{i}" for i in range(3)]

    def test_live_cib_required(self):
        self.config.env.set_cib_data("<cib />")
        self.env_assist.assert_raise_library_error(
            lambda: cluster.remove_nodes_from_cib(
                self.env_assist.get_env(),
                self.nodes,
            ),
            [
                fixture.error(
                    report_codes.LIVE_ENVIRONMENT_REQUIRED,
                    forbidden_options=["CIB"],
                )
            ],
            expected_in_processor=False,
        )

    def test_sucess_pcmk_running(self):
        self.config.services.is_running("pacemaker")
        for node in self.nodes:
            self.config.runner.pcmk.remove_node(
                node,
                name=f"remove_node.{node}",
            )
        cluster.remove_nodes_from_cib(self.env_assist.get_env(), self.nodes)

    def test_failure_pcmk_running(self):
        err_msg = "an error"
        self.config.services.is_running("pacemaker")
        self.config.runner.pcmk.remove_node(
            self.nodes[0],
        )
        self.config.runner.pcmk.remove_node(
            self.nodes[1],
            returncode=1,
            stderr=err_msg,
            name="remove_node_failure",
        )
        self.env_assist.assert_raise_library_error(
            lambda: cluster.remove_nodes_from_cib(
                self.env_assist.get_env(),
                self.nodes,
            ),
            [
                fixture.error(
                    report_codes.NODE_REMOVE_IN_PACEMAKER_FAILED,
                    node_list_to_remove=[self.nodes[1]],
                    node="",
                    reason=err_msg,
                )
            ],
            expected_in_processor=False,
        )

    def test_sucess_pcmk_not_running(self):
        cmd_env = dict(CIB_file=os.path.join(settings.cib_dir, "cib.xml"))
        self.config.services.is_running("pacemaker", return_value=False)
        for node in self.nodes:
            self.config.runner.place(
                [
                    settings.cibadmin,
                    "--delete-all",
                    "--force",
                    f"--xpath=/cib/configuration/nodes/node[@uname='{node}']",
                ],
                name=f"remove_node.{node}",
                env=cmd_env,
            )
        cluster.remove_nodes_from_cib(self.env_assist.get_env(), self.nodes)

    def test_failure_pcmk_not_running(self):
        err_msg = "an error"
        cmd_env = dict(CIB_file=os.path.join(settings.cib_dir, "cib.xml"))
        cmd = [settings.cibadmin, "--delete-all", "--force"]
        cmd_xpath = "--xpath=/cib/configuration/nodes/node[@uname='{}']"
        self.config.services.is_running("pacemaker", return_value=False)
        self.config.runner.place(
            cmd + [cmd_xpath.format(self.nodes[0])],
            name="remove_node_success",
            env=cmd_env,
        )
        self.config.runner.place(
            cmd + [cmd_xpath.format(self.nodes[1])],
            returncode=1,
            stderr=err_msg,
            name="remove_node_failure",
            env=cmd_env,
        )
        self.env_assist.assert_raise_library_error(
            lambda: cluster.remove_nodes_from_cib(
                self.env_assist.get_env(),
                self.nodes,
            ),
            [
                fixture.error(
                    report_codes.NODE_REMOVE_IN_PACEMAKER_FAILED,
                    node_list_to_remove=[self.nodes[1]],
                    node="",
                    reason=err_msg,
                )
            ],
            expected_in_processor=False,
        )



from pcs import settings
from pcs.common import reports
from pcs.common.reports.item import ReportItem
from pcs.common.str_tools import join_multilines
from pcs.lib.errors import LibraryError


def get_daemon_status(runner, name=None):
    cmd = [settings.booth_binary, "status"]
    if name:
        cmd += ["-c", name]
    stdout, stderr, return_value = runner.run(cmd)
    # 7 means that there is no booth instance running
    if return_value not in [0, 7]:
        raise LibraryError(
            ReportItem.error(
                reports.messages.BoothDaemonStatusError(
                    join_multilines([stderr, stdout])
                )
            )
        )
    return stdout


def get_tickets_status(runner, name=None):
    cmd = [settings.booth_binary, "list"]
    if name:
        cmd += ["-c", name]
    stdout, stderr, return_value = runner.run(cmd)
    if return_value != 0:
        raise LibraryError(
            ReportItem.error(
                reports.messages.BoothTicketStatusError(
                    join_multilines([stderr, stdout]),
                )
            )
        )
    return stdout


def get_peers_status(runner, name=None):
    cmd = [settings.booth_binary, "peers"]
    if name:
        cmd += ["-c", name]
    stdout, stderr, return_value = runner.run(cmd)
    if return_value != 0:
        raise LibraryError(
            ReportItem.error(
                reports.messages.BoothPeersStatusError(
                    join_multilines([stderr, stdout]),
                )
            )
        )
    return stdout



from unittest import mock, TestCase
from lxml import etree

from pcs_test.tools.assertions import assert_report_item_list_equal

from pcs.common.reports import ReportItemSeverity as severities
from pcs.common.reports import codes as report_codes
from pcs.lib.cib.node import PacemakerNode
from pcs.lib.cib.resource import remote_node


class FindNodeList(TestCase):
    def assert_nodes_equals(self, xml, expected_nodes):
        self.assertEqual(
            expected_nodes, remote_node.find_node_list(etree.fromstring(xml))
        )

    def test_find_multiple_nodes(self):
        self.assert_nodes_equals(
            """
            <resources>
                <primitive class="ocf" id="R1"
                    provider="pacemaker" type="remote"
                >
                    <instance_attributes>
                        <nvpair name="server" value="H1"/>
                    </instance_attributes>
                </primitive>
                <primitive class="ocf" id="R2"
                    provider="pacemaker" type="remote"
                >
                    <instance_attributes id="first-attribs">
                        <nvpair name="server" value="H2"/>
                    </instance_attributes>
                </primitive>
            </resources>
            """,
            [
                PacemakerNode("R1", "H1"),
                PacemakerNode("R2", "H2"),
            ],
        )

    def test_find_no_nodes(self):
        self.assert_nodes_equals(
            """
            <resources>
                <primitive class="ocf" id="D" provider="heartbeat" type="dummy">
                    <meta_attributes>
                        <nvpair name="server" value="H1"/>
                    </meta_attributes>
                </primitive>
            </resources>
            """,
            [],
        )

    def test_find_nodes_without_server(self):
        self.assert_nodes_equals(
            """
            <resources>
                <primitive class="ocf" id="R1"
                    provider="pacemaker" type="remote"
                >
                </primitive>
            </resources>
            """,
            [
                PacemakerNode("R1", "R1"),
            ],
        )

    def test_find_nodes_with_empty_server(self):
        # it does not work, but the node "R1" is visible as remote node in the
        # status
        self.assert_nodes_equals(
            """
            <resources>
                <primitive class="ocf" id="R1"
                    provider="pacemaker" type="remote"
                >
                    <instance_attributes id="first-attribs">
                        <nvpair name="server" value=""/>
                    </instance_attributes>
                </primitive>
            </resources>
            """,
            [
                PacemakerNode("R1", "R1"),
            ],
        )


class FindNodeResources(TestCase):
    def assert_resources_equals(self, node_identifier, xml, resource_id_list):
        self.assertEqual(
            resource_id_list,
            [
                resource_element.attrib["id"]
                for resource_element in remote_node.find_node_resources(
                    etree.fromstring(xml), node_identifier
                )
            ],
        )

    def test_find_all_resources(self):
        self.assert_resources_equals(
            "HOST",
            """<resources>
                <primitive class="ocf" id="R1"
                    provider="pacemaker" type="remote"
                >
                    <instance_attributes>
                        <nvpair name="server" value="HOST"/>
                    </instance_attributes>
                </primitive>
                <primitive class="ocf" id="R2"
                    provider="pacemaker" type="remote"
                >
                    <instance_attributes id="first-attribs">
                        <nvpair name="server" value="HOST"/>
                    </instance_attributes>
                </primitive>
            </resources>""",
            ["R1", "R2"],
        )

    def test_find_by_resource_id(self):
        self.assert_resources_equals(
            "HOST",
            """<resources>
                <primitive class="ocf" id="HOST"
                    provider="pacemaker" type="remote"
                />
            </resources>""",
            ["HOST"],
        )

    def test_ignore_non_remote_primitives(self):
        self.assert_resources_equals(
            "HOST",
            """<resources>
                <primitive class="ocf" id="HOST"
                    provider="heartbeat" type="Dummy"
                />
            </resources>""",
            [],
        )


class GetNodeNameFromResource(TestCase):
    def test_return_name(self):
        self.assertEqual(
            "R",
            remote_node.get_node_name_from_resource(
                etree.fromstring(
                    """
                <primitive class="ocf" id="R" provider="pacemaker" type="remote"
                />
            """
                )
            ),
        )

    def test_return_name_ignore_host(self):
        self.assertEqual(
            "R",
            remote_node.get_node_name_from_resource(
                etree.fromstring(
                    """
                <primitive class="ocf" id="R" provider="pacemaker" type="remote"
                >
                    <instance_attributes>
                        <nvpair name="server" value="HOST"/>
                    </instance_attributes>
                </primitive>
            """
                )
            ),
        )

    def test_return_none_when_primitive_is_without_agent(self):
        case_list = [
            '<primitive id="R"/>',
            '<primitive id="R" class="ocf"/>',
            '<primitive id="R" class="ocf" provider="pacemaker"/>',
        ]
        for case in case_list:
            self.assertIsNone(
                remote_node.get_node_name_from_resource(etree.fromstring(case)),
                "for '{0}' is not returned None".format(case),
            )

    def test_return_none_when_primitive_is_not_pacemaker_remote(self):
        self.assertIsNone(
            remote_node.get_node_name_from_resource(
                etree.fromstring(
                    """
                <primitive class="ocf" id="R" provider="heartbeat" type="dummy"
                />
            """
                )
            )
        )


class Validate(TestCase):
    @staticmethod
    def validate(
        instance_attributes=None, node_name="NODE-NAME", host="node-host"
    ):
        existing_nodes_names = ["R"]
        existing_nodes_addrs = ["RING0", "RING1"]
        resource_agent = mock.MagicMock()
        return remote_node.validate_create(
            existing_nodes_names,
            existing_nodes_addrs,
            resource_agent,
            node_name,
            host,
            instance_attributes if instance_attributes else {},
        )

    def test_report_conflict_node_name(self):
        assert_report_item_list_equal(
            self.validate(
                node_name="R",
                host="host",
            ),
            [
                (
                    severities.ERROR,
                    report_codes.ID_ALREADY_EXISTS,
                    {
                        "id": "R",
                    },
                    None,
                )
            ],
        )

    def test_report_conflict_node_host(self):
        assert_report_item_list_equal(
            self.validate(
                host="RING0",
            ),
            [
                (
                    severities.ERROR,
                    report_codes.ID_ALREADY_EXISTS,
                    {
                        "id": "RING0",
                    },
                    None,
                )
            ],
        )

    def test_report_conflict_node_host_ring1(self):
        assert_report_item_list_equal(
            self.validate(
                host="RING1",
            ),
            [
                (
                    severities.ERROR,
                    report_codes.ID_ALREADY_EXISTS,
                    {
                        "id": "RING1",
                    },
                    None,
                )
            ],
        )

    def test_report_used_disallowed_server(self):
        assert_report_item_list_equal(
            self.validate(instance_attributes={"server": "A"}),
            [
                (
                    severities.ERROR,
                    report_codes.INVALID_OPTIONS,
                    {
                        "option_type": "resource",
                        "option_names": ["server"],
                        "allowed": [],
                        "allowed_patterns": [],
                    },
                    None,
                )
            ],
        )
